# frozen_string_literal: true

module Gdas::ArtifactAnalyzers::JobSets

  GDAS_PROC_JOB_SIZE = 10000

  # An AssetJobSet is responsible for collecting resources and generating fanout jobs for this
  # account and asset type combination. This is a parent class that contains shared functionality
  # for the two implementations below that inherit from it.
  #
  # An implementation of an asset job set should expose the following methods:
  #
  # has_resource?(resource_id)
  #  - Returns boolean indicating whether the job set currently has the resource associated with the
  #    given ID
  #
  # get_resource(resource_id)
  #  - Returns the resource data associated with the given resource ID
  #
  # get_fanout_jobs!(first_bill_time, last_bill_time)
  #  - Returns all processing jobs that are ready for fanout. Deletes all resource data when finished
  #    to save memory
  #
  # new_resource(resource_data, partition value)
  #  - Adds a new resource to the job set.
  #  - Returns a boolean indicating whether there are any processing jobs that are ready to be
  #    fanned out
  #
  # finished_product(processed_prods)
  #  - Indicates to the job set that a product has been fully processed.
  class AssetJobSet

    # Constructor for an AssetJobSet
    #
    # @account
    #   - The AWS account owner ID for this job set
    #
    # @asset_type
    #   - The asset type that is being collected in this job set
    #
    # @resources
    #   - A hash mapping resource ID to resource data
    #
    # @job_id_sets
    #   - A hash mapping [partition_lower_bound, partition_upper_bound] pairs to a Set containing all of the
    #     resource IDs with partition values within this range.
    #   - Resource ID sets should only be added to this hash when they are complete and ready to
    #     be fanned out.
    #
    # @partition_val_min
    #   - The minimum partition value in the partition value series
    #
    # @partition_val_max
    #   - The maximum partition value in the partition value series
    #
    # @partition_val_increment_func
    #   - A function that when passed a partition value, increments it to the next value in the series. If the passed
    #     value is equal to the maximum value in the series, the same value is returned
    def initialize(owner_account, asset_type, partition_val_min, partition_val_max, partition_val_increment_func)
      @account = owner_account
      @asset_type = asset_type
      @resources = {}
      @job_id_sets = {}
      @partition_val_min = partition_val_min
      @partition_val_max = partition_val_max
      @partition_val_increment_func = partition_val_increment_func
    end

    def get_resource(resource_id)
      @resources[resource_id]
    end

    def has_resource?(resource_id)
      @resources.key?(resource_id)
    end

    # Generates processing job payloads that are ready to be fanned out. Deletes resources
    # from the resource hash as the jobs are generated in order to save memory.
    #
    # Assumes that resource ID sets in the @job_id_sets hash are complete and ready to be fanned out
    #
    # Returns jobs of the following structure:
    # {
    #   account_id: ...
    #   asset_type: ...
    #   first_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
    #   last_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
    #   partition_lower_bound: String, # Partition value for lower bound of job,
    #                                    used for cache building in processing
    #   partition_upper_bound: String, # Partition value for upper bound of job,
    #                                    used for cache building in processing
    #   resources: [
    #     {
    #       resource_id: String,
    #       cloud_account_id: String,
    #       cloud: String,
    #       region: String,
    #       asset_type: String,
    #       start_time: DateTime,
    #       most_recently_found_time: DateTime
    #       tags: {tag_key: tag_value},
    #       custom_columns: {ColumnName: column_value}
    #     },
    #     ...
    #   ]
    # }
    def get_fanout_jobs!(first_bill_time, last_bill_time, cloud_name)
      fanout_jobs = []
      @job_id_sets.keys.each do |bound_pair|
        id_set = @job_id_sets.delete(bound_pair)
        job_resources = []
        if !id_set.nil? && !id_set.empty?
          id_set.each { |res_id| job_resources << @resources.delete(res_id) }
        end
        fanout_jobs << {
            account_id: @account,
            asset_type: @asset_type,
            cloud: cloud_name,
            first_bill_time: first_bill_time,
            last_bill_time: last_bill_time,
            partition_lower_bound: bound_pair[0],
            partition_upper_bound: bound_pair[1],
            resources: job_resources
        }
      end
      fanout_jobs
    end

    def new_resource(resource_data, partition_val)
      raise NotImplementedError
    end

    def finished_product(proc_prods)
      raise NotImplementedError
    end

    def generate_empty_job_set
      raise NotImplementedError
    end
  end
end
