# frozen_string_literal: true
require 'gdas/artifact_analyzers/asset_job_sets/asset_job_set'

module Gdas::ArtifactAnalyzers::JobSets

  # An implementation of an asset job set that assumes there is more than one product associated
  # with the given asset type in the xform artifact. Due to this, we cannot fanout processing job
  # chunks until all products associated with this asset type have been fully processed.
  # This is because there could be the same resource listed twice in the xform artifact under
  # different product names.
  class MultiProductAssetJobSet < AssetJobSet

    # Constructor for a MultiProductAssetJobSet
    #
    # @associated_products
    #   - All of the products associated with the given asset type that are present in the artifact
    #
    # @partition_val_set
    #   - A Set containing all of the partition values found in the artifact for this asset type
    #
    # @partition_val_to_id
    #   - a hash mapping partition value to the associated resource ID
    def initialize(owner_account, asset_type, associated_products, partition_val_min, partition_val_max, partition_val_inc_func)
      super(owner_account, asset_type, partition_val_min, partition_val_max, partition_val_inc_func)
      @associated_products = associated_products
      @partition_val_set = Set.new
      @partition_val_to_id = {}
    end

    # Adds a new resource to the running collection. Since this asset type has multiple products
    # associated with it, we cannot incrementally fanout so always return false
    def new_resource(resource_data, partition_val)
      res_id = resource_data[:resource_id]
      @resources[res_id] = resource_data
      @partition_val_set.add(partition_val)
      @partition_val_to_id[partition_val] ||= res_id
      false # Multi-product assets cannot fanout until all associated products have been processed
    end

    # Indicates to this job set that a product has finished processing. If all products associated
    # with this asset have been processed, then generate the processing jobs
    def finished_product(proc_prods)
      generate_jobs if done_processing?(proc_prods)
    end

    def generate_empty_job_set
      @partition_val_set.add(@partition_val_min)
      generate_jobs(true) # Force job through with empty resources
    end

    private

    # Should only fanout if all products associated with the given asset type have either been
    # processed or do not appear in the xform artifact
    def done_processing?(proc_prods)
      @associated_products.all? { |p| proc_prods.include?(p) }
    end

    # Determines the partition value cutoffs for each job chunk and adds the resource ID sets to the
    # @job_id_sets hash for fanout
    def generate_jobs(allow_empty_resources = false)
      # return if partition_val_set is empty OR if there are no resources and we are not a empty product termination job. 
      return if @partition_val_set.empty? || (@resources.empty? && !allow_empty_resources)
      sorted_partition_vals = @partition_val_set.to_a.sort
      num_partitions = (sorted_partition_vals.size.to_f / GDAS_PROC_JOB_SIZE).ceil
      partition_val_lower = @partition_val_min

      sorted_partition_vals.each_slice(GDAS_PROC_JOB_SIZE).with_index do |partition, index|
        # If this is the last chunk then the upper bound is the max partition value, else the largest partition value
        # in this chunk
        partition_val_upper = (index == num_partitions - 1) ? @partition_val_max : partition[-1]

        # Partition value lower and upper bound pair for this chunk
        bounds = [partition_val_lower, partition_val_upper]

        # Create the set of resource IDs to be used for fanout job from this chunk of partition values
        @job_id_sets[bounds] = Set.new(partition.map { |partition_val| @partition_val_to_id[partition_val] }.compact)
        partition_val_lower = @partition_val_increment_func.call(partition[-1])
      end
    end
  end
end
