# frozen_string_literal: true
require 'gdas/artifact_analyzers/asset_job_sets/asset_job_set'

module Gdas::ArtifactAnalyzers::JobSets

  # An implementation of an asset job set that relies on only one product associated with the given
  # asset type is present in the xform artifact.
  # If only one product for this asset type is present on the xform artifact, then we know that
  # there will be exactly one row per resource in the xform artifact and they will be sorted by
  # the partition column.
  # This means that we can safely generate processing jobs before finishing processing the
  # associated product.
  class OneProductAssetJobSet < AssetJobSet

    # Constructor for a OneProductAssetJobSet
    #
    # @associated_product
    #   - The single product name present in the artifact associated with the given asset type
    #
    # @cur_partition_lower
    #   - The partition value lower bound value for the chunk currently being collected
    #
    # @cur_partition_upper
    #   - The partition value upper bound value for the chunk currently being collected
    #
    # @cur_job_res_ids
    #   - A Set of the resource IDs that are part of the chunk currently being collected.
    #     When this set of resource IDs is complete, the set will be moved to the @job_id_sets hash
    def initialize(owner_account, asset_type, associated_product, partition_val_min, partition_val_max, partition_val_inc_func)
      super(owner_account, asset_type, partition_val_min, partition_val_max, partition_val_inc_func)
      @associated_product = associated_product
      @cur_partition_lower = @partition_val_min
      @cur_partition_upper = @partition_val_min
      @cur_job_res_ids = Set.new
    end

    # Adds a new resource to this asset job set and moves the partition value upper bound to the given partition value.
    # If the current set of resource IDs has reached the maximum size, it will move it to the complete ID sets hash to
    # be retrieved for fanout. Returns a boolean indicating whether there are any jobs that are complete and ready to
    # be collected.
    def new_resource(resource_data, partition_val)
      finished_job if @cur_job_res_ids.size >= GDAS_PROC_JOB_SIZE

      @cur_partition_upper = partition_val
      res_id = resource_data[:resource_id]
      @resources[res_id] = resource_data
      @cur_job_res_ids.add(res_id)
      @job_id_sets.size > 0 # Fanout is ready if there is at least one job in the job id sets hash
    end

    # Indicates to this job set that a product has been fully processed. If the product associated
    # with this asset type is in the Set of fully processed products, and there were resources
    # processed for this asset type, then all remaining resources will be marked as ready to fanout
    def finished_product(proc_prods)
      if proc_prods.include?(@associated_product) && (@cur_job_res_ids.size > 0 || @cur_partition_lower != @partition_val_min)
        @job_id_sets[[@cur_partition_lower, @partition_val_max]] = @cur_job_res_ids
        @cur_job_res_ids = Set.new
        @cur_partition_lower = @partition_val_min
        @cur_partition_upper = @partition_val_min
      end
    end

    def generate_empty_job_set
      @job_id_sets[[@cur_partition_lower, @partition_val_max]] = nil
    end

    private

    # Moves the current set of resource IDs to the @job_id_sets hash with the current partition value bounds
    # and resets the bounds to the next chunk to be processed.
    def finished_job
      @job_id_sets[[@cur_partition_lower, @cur_partition_upper]] = @cur_job_res_ids
      @cur_job_res_ids = Set.new
      next_partition_val = @partition_val_increment_func.call(@cur_partition_upper)
      @cur_partition_lower = next_partition_val
      @cur_partition_upper = next_partition_val
    end
  end
end
