# frozen_string_literal: true
#
module Gdas::ArtifactAnalyzers::Aws

  class AwsXformAnalyzer < ::Gdas::ArtifactAnalyzers::XformAnalyzer

    PRODUCTS_WITHOUT_ACCT_IN_ARN = [
      'AWS Import/Export Snowball',
      'Amazon Route 53',
      'Amazon API Gateway',
      'Amazon Bedrock',
      "Amazon Security Lake"
    ].freeze

    # Known products that list entries in the bill with timestamps that are in the future
    # These products will be ignored when it comes to determining global last bill time which is
    # used for determining whether assets are active or not
    BILL_TIME_BLACKLIST = [
      'Savings Plans for AWS Compute usage'
    ].freeze

    class << self
      # Names of the x-form artifacts that this analyzer supports
      def supported_artifacts
        ['gdas_artifact_with_metadata', 'gdas_ec2_instances', 'gdas_rds_instances', 'gdas_ec2_volumes_and_snapshots']
      end

      def cloud_name
        Gdas::Clouds::AwsCloudConfig::CLOUD_NAME
      end

      def column_cloud_account_id
        'LinkedAccountId'
      end

      def should_create_termination_jobs?
        true
      end
    end

    private

    # For AWS artifacts cloud_account_id is used as a sort parameter. We can fan
    # out jobs when the line item's account changes.
    def handle_artifact_transitions(row)
      product_name = @stream_adapter.column_value_for_row(row, self.class.required_columns[:product_name])
      account_id = @stream_adapter.column_value_for_row(row, self.class.required_columns[:cloud_account_id])

      if @current_product.nil?
        # This is the first data row, initialize and set current account/product and continue
        @current_product = product_name
        @current_account = account_id
        @products_processed_by_account = {}
      elsif @current_product != product_name || @current_account != account_id
        finish_product
        @fanout_jobs_ready = true
        @current_product = product_name
        @current_account = account_id
      end
    end

    def finish_product
      # First time finishing a product for an account
      if !@products_processed_by_account[@current_account]
        @products_processed_by_account[@current_account] = Set.new
      end

      # Short circuit if the artifact does not seem to be ordered correctly.
      if @products_processed_by_account[@current_account].include?(@current_product)
        message = 'A product was marked as completed more than once for an account'
        logger.error "#{message} Path: #{@stream_adapter.file_path}, Customer: #{@stream_adapter.customer_id}, Product: #{@current_product}, Account: #{@current_account}"
        alert.error(
          message,
          s3_path: @stream_adapter.file_path,
          customer_id: @stream_adapter.customer_id,
          current_account: @current_account,
          current_product: @current_product,
          cloud: self.class.cloud_name
        )
        raise ::Platform::NonRetryableError, orig_exception: message
      end

      @products_processed_by_account[@current_account].add(@current_product)

      if @job_sets[@current_account]
        @job_sets[@current_account].values.each do |job_set|
          job_set.finished_product(@products_processed_by_account[@current_account])
        end
      end
    end

    # For some GovCloud assets, the account ID column does not match the account ID in the ARN.
    # While we want the G-DAS to notice and skip assets that have account mismatches, we are choosing to ignore
    # the mismatch in the case of GovCloud ARNs because this is an expected case. See ENG-37151.
    def ignore_resource?(resource_id, account_id, product_name)
      # Note: +account_id+ is the owner ID from the +LinkedAccountId+ column.

      should_skip =
        resource_id.start_with?('arn:') && # It's an ARN
        !resource_id.start_with?('arn:aws-us-gov:') && # but not for GovCloud,
        !PRODUCTS_WITHOUT_ACCT_IN_ARN.any? {|p| product_name.include? p } && # and we expect that the ARN includes the owner ID,
        !resource_id.include?(":#{account_id}:") # and yet the ARN doesn't include the owner ID.

      if should_skip
        tags = %W(
            product_name:#{product_name}
            cloud:#{self.class.cloud_name}
        )
        logger.debug(
          'Found conflicting account ID. Skipping single occurrence of '\
          "resource ID: #{resource_id}. Expected account: #{account_id}."
        )
        statsd.increment('xform_analyzer.account_conflict_row_skipped', tags: tags)
      end

      should_skip
    end

    def ignore_bill_time?(product_name)
      BILL_TIME_BLACKLIST.include?(product_name)
    end

    # Return the prefix and prefix length to filter tags by. Only tags that pass the filter are added to the job.
    # In the bill, user-defined tag keys start with the prefix 'user:'.
    def filter_tag_key_prefix
      return 'user:', 5
    end
  end
end
