# frozen_string_literal: true

require 'gdas/assets/generic_derived_asset_scanner/context'
require 'gdas/assets/generic_derived_asset_scanner/csv_stream/csv_analyzer_utilities'
require 'gdas/artifact_analyzers/asset_job_sets/asset_job_set_factory'
require 'time'
require 'memoist'


module Gdas::ArtifactAnalyzers

  # Cloud specific subclasses of the XformAnalyzer need to implement the the following methods at minimum:
  # - +cloud_name+  specify what cloud provider this analyzer is for
  # - +supported_artifacts+  specify names of the xform artifacts this analyzer can be used with
  # - +column_cloud_account_id+
  #
  # These other methods can also be optionally overridden for cloud specific logic:
  # - +required_columns+
  # - +ignore_resource?+  to filter out a resource from the generated jobs
  # - +ignore_bill_time?+  to filter out products from determining global bill timestamps
  # - +add_resource_data+ and +update_resource_data+  to add additional fields to the payload
  # - +filter_tag_key_prefix+  use the prefix of tag keys to determine what tags make it into the payload
  class XformAnalyzer
    extend Gdas::Assets::GenericDerivedAssetScanner::CsvStream::CsvAnalyzerUtilities
    include Gdas::Assets::GenericDerivedAssetScanner::Context
    include Import["logger", "alert", "statsd"]

    LINKED_ACCOUNT_META_DATA_ROW_TYPE = -1
    ASSET_META_DATA_ROW_TYPE = 0
    DATA_ROW_TYPE = 1

    class << self
      extend ::Memoist

      # Constants for the Xform column names
      def required_columns
        {
          resource_id: 'ResourceId',
          resource_id_sha1: 'ResourceIdSha1',
          product_name: 'ProductName',
          cloud_account_id: column_cloud_account_id,
          region: 'Region',
          start_time: 'StartTime',
          most_recently_found_time: 'MostRecentlyFoundTime',
          tags: 'Tags',
          row_type: 'RowType'
        }.freeze
      end
      memoize :required_columns

      def column_cloud_account_id
        raise NotImplementedError
      end

      def registry
        ::Gdas::Assets::GenericDerivedAssetScanner::Registry
      end

      # This is used to determine whether or not we should look at a Metadata row 
      # to determine if we need to perform account clean up job(s).
      # Ex: Asset for Account 12345 has data in bill 04-24 but has no data for Account 12345 in 05-24. 
      # This may leave stranded assets we need to clean up. 
      # @linked_visited_accounts_and_assets variable needs be initialized as {} if should_create_termination_jobs? is true
      # Make sure that get_accounts_and_assets_visited, get_accounts_for_billing_artifact, get_accounts_visited are implemented.
      def should_create_termination_jobs?
        false
      end

      def cloud_name
        raise NotImplementedError
      end

      # Names of the x-form artifacts that this analyzer supports
      def supported_artifacts
        raise NotImplementedError
      end

      # Override this to add additional fields to a resource in the job. This function can be used to add other artifact
      # fields that are standard to a cloud, without having to add them to the custom_columns in the payload. These
      # fields should also be added to +required_columns+ so as not to be included as a custom_column.
      def add_resource_data(resource_data, stream_adapter, row) end

      # Update a resource that was already in the job set with any additional columns needed
      def update_resource_data(stored_resource, stream_adapter, row) end
    end

    def initialize(stream_adapter, artifact_name, *args)
      @stream_adapter = stream_adapter
      @skipped_resource_count = 0
      @used_resource_count = 0
      @first_bill_time = nil
      @last_bill_time = nil
      @products_in_artifact = Set.new
      @products_processed = Set.new
      @current_product = nil
      @fanout_jobs_ready = false # A flag to determine when to attempt retrieving fanout jobs
      @job_sets = {}
      @artifact_name = artifact_name
      @linked_accounts = Set.new
      @linked_visited_accounts_and_assets = {}

      super(*args) # Import
    end

    def get_accounts_for_billing_artifact
      @linked_accounts
    end

    def get_accounts_visited
      @linked_visited_accounts_and_assets.keys
    end

    def get_accounts_and_assets_visited
      @linked_visited_accounts_and_assets
    end

    # Stream and filter a zipped, xform artifact file from S3, and yield jobs that
    # can be fanned out to processing.
    #
    # Params:
    # - +stream_adapter+:: S3StreamFilterAdapter to access the stream.
    #
    # Yields:
    # {
    #   account_id: ...
    #   asset_type: ...
    #   cloud: String, # the cloud service this job is associated with
    #   first_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
    #   last_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
    #   partition_lower_bound: String, # partition value for lower bound of job,
    #                               used for cache building in processing. Inclusive range
    #   partition_upper_bound: String, # partition value for upper bound of job,
    #                               used for cache building in processing. Inclusive range
    #   resources: [
    #     {
    #       resource_id: String,
    #       cloud_account_id: String,
    #       region: String,
    #       asset_type: String,
    #       start_time: DateTime,
    #       most_recently_found_time: DateTime
    #       tags: {tag_key: tag_value},
    #       custom_columns: {ColumnName: column_value}
    #     },
    #     ...
    #   ]
    # }
    def each_fanout_job
      # A hash from account ID and asset type to an AssetJobSet object which is responsible
      # for managing the resources associated with that asset type and generating fanout
      # jobs when ready. It has the following structure:
      #
      # {
      #   "cloud_account_id" =>
      #     {
      #       ['service', 'asset_type'] => AssetJobSet
      #     }
      # }

      # Set up the stream and process the headers.
      @stream_adapter.start_stream

      # Defensively check that the columns we need exist in the header.
      enforce_required_columns(@stream_adapter)

      reached_first_data_row = false
      @stream_adapter.data_rows_each do |row|
        row_type = @stream_adapter.column_value_for_row(row, self.class.required_columns[:row_type]).to_i

        # Products that are not part of the G-DAS registry are filtered out by the S3 file streamer
        # so do not affect the global first and last bill times
        # Xform artifacts are sorted by row type so it is guaranteed that all metadata rows come
        # before the data rows
        if row_type == ASSET_META_DATA_ROW_TYPE
          process_asset_metadata_row(row)
        elsif row_type == LINKED_ACCOUNT_META_DATA_ROW_TYPE
          process_account_metadata_row(row)
        else
          # @pnames_to_asset_info and @asset_types_to_pnames (we want initialize them only once when we are done running metadata)
          reached_first_data_row = initialize_asset_type_lookups unless reached_first_data_row

          handle_artifact_transitions(row)

          # Only attempt to retrieve fanout jobs from AssetJobSets if at least one of them has indicated
          # that there are fanout jobs ready
          if @fanout_jobs_ready
            get_fanout_jobs!.each { |job| yield job }
            @fanout_jobs_ready = false
          end

          process_data_row(row)
        end
      end

      logger.warn do
        "Identified #{@used_resource_count} unique, usable resources and skipped #{@skipped_resource_count} "\
          "resource IDs for customer #{@stream_adapter.customer_id} in xform artifact path '#{@stream_adapter.file_path}'"
      end

      if @used_resource_count != 0
        finish_product
      end

      # For Accounts in the Bill where there are no usage for the given Product line.
      create_missing_jobs_for_artifact
      get_fanout_jobs!.each { |job| yield job }
    end

    private

   # For Accounts in the Bill where there are no usage for the given Product line. 
   # Make sure to create a job for each product and account so we correctly emit a termination fanout. 
    def create_missing_jobs_for_artifact
      if self.class.should_create_termination_jobs?
        begin 
          assets_in_product_with_jobs = @products_in_artifact.map{ |product_name| @pnames_to_asset_info[product_name] }
            .compact
            .flatten
            .to_set

          get_accounts_for_billing_artifact.each do |account_id|
            missing_assets_for_account = get_missing_assets_for_account(assets_in_product_with_jobs, account_id)
            missing_assets_for_account.each do |asset_info|
              asset_type = asset_info[:asset_type]
              asset_collector = asset_info[:collector]
              if asset_collector.gdas_artifact_name != @artifact_name
                # If this is not the artifact we expect to see resources in for this product skip. 
                next
              end
  
              # Create keys in job_sets hash if they do not already exist
              build_job_set_hash(account_id.to_s, asset_type, asset_collector)
              job_set = @job_sets[account_id.to_s][asset_type]
              job_set.generate_empty_job_set
  
              tags = %W(
                customer_id: #{@stream_adapter.customer_id}
                asset_type: #{asset_type}
              )
              statsd.increment('gdas.assets.missing_termination_job', tags: tags)
            end
          end
        rescue => ex
          message = 'Error creating termination jobs.'
          logger.error "#{message} Path: #{@stream_adapter.file_path}, Customer: #{@stream_adapter.customer_id}, ex: #{ex.message}"
          alert.exception(
            ex,
            message: message,
            s3_path: @stream_adapter.file_path,
            customer_id: @stream_adapter.customer_id,
            cloud: self.class.cloud_name
          )
          tags = %W(
            customer_id: #{@stream_adapter.customer_id}
            cloud: self.class.cloud_name
            error: true
          )
          statsd.increment('gdas.assets.missing_account_jobs', tags: tags)
          # If there are any errors we don't need to fail anything.
        end
      end 
    end

    def get_missing_assets_for_account(assets_in_product_with_jobs, account_id)
      jobs = get_accounts_and_assets_visited[account_id]
      # If there are no jobs associated with an account - mark as empty set.
      jobs = Set.new if jobs.nil?
      # All Assets in Artifact
      if jobs.size >= assets_in_product_with_jobs.size
        # We have enqueued a job for every product-type in this artifact for this account continue on!
        []
      else
        # We need to only do this for Products + Accounts we've not seen.
        assets_in_product_with_jobs - jobs
      end
    end

    # Metadata rows occur before data rows in the artifact. These rows are used
    # to identify the list of products found in the artifact as well as the
    # first and last bill timestamp.
    def process_asset_metadata_row(row)
      product_name = @stream_adapter.column_value_for_row(row, self.class.required_columns[:product_name])
      @products_in_artifact.add(product_name)

      unless ignore_bill_time?(product_name)
        first_time, last_time = parse_bill_times(@stream_adapter, row)
        @first_bill_time = first_time if @first_bill_time.nil? || (first_time < @first_bill_time)
        @last_bill_time = last_time if @last_bill_time.nil? || (last_time > @last_bill_time)
      end
    end

    def process_account_metadata_row(row)
      begin
        if self.class.should_create_termination_jobs?
          combined_ids = @stream_adapter.column_value_for_row(row, self.class.required_columns[:cloud_account_id])
          account_ids = combined_ids.split(";").map{|x| x.to_i}
          @linked_accounts = account_ids.to_set  
        end   
      rescue => ex
        message = 'Error parsing Account Metadata row.'
        logger.error "#{message} Path: #{@stream_adapter.file_path}, Customer: #{@stream_adapter.customer_id}, ex: #{ex.message}"
        alert.exception(
          ex,
          message: message,
          s3_path: @stream_adapter.file_path,
          customer_id: @stream_adapter.customer_id,
          cloud: self.class.cloud_name
        )
        tags = %W(
          customer_id: #{@stream_adapter.customer_id}
          cloud: self.class.cloud_name
          error: true
        )
        statsd.increment('gdas.assets.account_metadata_parse', tags: tags)
      end
      # This is fine we can continue on - we do not need to fail for this. 
    end

    def build_job_set_hash(account_id, asset_type, asset_collector)
      # Create keys in job_sets hash if they do not already exist
      @job_sets[account_id] ||= {}
      @job_sets[account_id][asset_type] ||= JobSets::AssetJobSetFactory.new_asset_job_set(
        account_id,
        asset_type,
        @products_in_artifact,
        @asset_types_to_pnames,
        asset_collector.partition_val_min,
        asset_collector.partition_val_max,
        -> (val) { asset_collector.increment_partition_val(val) }
      )
    end

    # This function processes a data row from the artifact. The data from the
    # given row creates or updates data in the @job_sets data object.
    def process_data_row(row)
      product_name = @stream_adapter.column_value_for_row(row, self.class.required_columns[:product_name])
      account_id = @stream_adapter.column_value_for_row(row, self.class.required_columns[:cloud_account_id])
      resource_id = @stream_adapter.column_value_for_row(row, self.class.required_columns[:resource_id])

      # Should never happen because xform query filters by null and empty resource IDs
      if resource_id.blank?
        message = 'Resource with blank resource ID found.'
        logger.warn "#{message} Path: #{@stream_adapter.file_path}, Customer: #{@stream_adapter.customer_id}"
        alert.warn(
          message,
          s3_path: @stream_adapter.file_path,
          customer_id: @stream_adapter.customer_id,
          cloud: self.class.cloud_name
        )
        return
      end

      return if ignore_resource?(resource_id, account_id, product_name)

      asset_info = lookup_asset_info(product_name, resource_id)

      # If the asset type is nil, it is not supported. Increment skipped count and skip this resource
      if asset_info.nil?
        @skipped_resource_count += 1
        return
      end

      asset_type = asset_info[:asset_type]
      asset_collector = asset_info[:collector]
      # Create keys in job_sets hash if they do not already exist
      build_job_set_hash(account_id, asset_type, asset_collector)

      start_time = @stream_adapter.column_value_for_row(row, self.class.required_columns[:start_time])
      most_recently_found_time = @stream_adapter.column_value_for_row(row, self.class.required_columns[:most_recently_found_time])

      job_set = @job_sets[account_id][asset_type]

      custom_column_names = @stream_adapter.column_headers - self.class.required_columns.values
      custom_column_vals = custom_column_names.each.with_object({}) do |col_name, h|
        h[col_name] = @stream_adapter.column_value_for_row(row, col_name)
      end

      # If we've gotten this far we will be triggering a fanout job for this account so we need to mark 
      # this account as "visited"/"seen" in this billing artifact. 
      if self.class.should_create_termination_jobs?
        # This variable will be initialized if should_create_termination_jobs? is true
        @linked_visited_accounts_and_assets[account_id.to_i] ||= Set.new 
        @linked_visited_accounts_and_assets[account_id.to_i] << asset_info
      end

      # If we have already discovered this resource for the same g-das asset type, but different
      # product name, then update the values
      if job_set.has_resource?(resource_id)
        stored_resource = job_set.get_resource(resource_id)
        stored_resource[:start_time] = start_time if start_time < stored_resource[:start_time]
        if most_recently_found_time > stored_resource[:most_recently_found_time]
          stored_resource[:most_recently_found_time] = most_recently_found_time
          stored_resource[:tags] = parse_tags(row)
          stored_resource[:custom_columns] = custom_column_vals
          self.class.update_resource_data(stored_resource, @stream_adapter, row)
        end
      else
        partition_val = @stream_adapter.column_value_for_row(row, asset_collector.job_partition_column)


        # Add new resource to output hash
        resource_data = {
          resource_id: resource_id,
          job_partition_value: partition_val,
          cloud_account_id: account_id,
          region: @stream_adapter.column_value_for_row(row, self.class.required_columns[:region]),
          asset_type: asset_type,
          start_time: start_time,
          most_recently_found_time: most_recently_found_time,
          tags: parse_tags(row),
          custom_columns: custom_column_vals
        }
        self.class.add_resource_data(resource_data, @stream_adapter, row)
        @fanout_jobs_ready = job_set.new_resource(resource_data, partition_val)
        @used_resource_count += 1
      end
    end

    # The artifact is sorted by some parameters. This method leverages changes
    # in those columns to modify the analyzers state.
    #
    # For example, if a row has a different product than the previous row we can
    # assume there are no more rows for the previous product and call
    # finish_product.
    def handle_artifact_transitions(row)
      product_name = @stream_adapter.column_value_for_row(row, self.class.required_columns[:product_name])

      if @current_product.nil?
        # This is the first data row, set the current product and continue
        @current_product = product_name
      elsif @current_product != product_name
        finish_product
        @fanout_jobs_ready = true
        @current_product = product_name
      end
    end

    # Initializes two instance variables for efficient lookups of the derived asset registry.
    #
    # The first is a mapping from product name to asset type, regular expression for
    # the resource ID, and the registered collector class. Note: Product names (the key) have to be unique, but there
    # may be multiple product names pointing to the same asset type.
    #
    # The second is a mapping from asset type to the list of product names associated with that
    # asset type
    #
    # Sets pnames_to_asset_info to:
    # - Hash lookup of product name to asset type definitions. Structure:
    #     {
    #       product_name: [
    #         {
    #           resource_id_pattern: /regular-expression/, # or +nil+
    #           asset_type: ['sqs', 'queue'],
    #           collector: collector class
    #         }
    #       ]
    #     }
    #
    # and asset_types_to_pnames to:
    # - Hash lookup of asset types to product names associated with that asset type. Structure:
    #     {
    #       asset_type: [product_name1, product_name2, ...]
    #     }
    def initialize_asset_type_lookups
      @asset_types_to_pnames = {}
      @pnames_to_asset_info = {}
      self.class.registry.instance.each do |vals, collector|
        # Only collect asset types that are registered with the current artifact
        next unless collector.gdas_artifact_name == @artifact_name && vals[:cloud] == self.class.cloud_name
        asset_type = [vals[:service], vals[:asset_type]]

        # Get list of all product names that are matching in `@products_in_artifact`
        if vals[:product_match_name]
          vals[:product_name].concat(@products_in_artifact.select { |p| p =~ vals[:product_match_name] })
        end

        @asset_types_to_pnames[asset_type] = vals[:product_name]

        vals[:product_name].each do |product_name|
          unless @pnames_to_asset_info.key?(product_name)
            @pnames_to_asset_info[product_name] = []
          end

          @pnames_to_asset_info[product_name] << {
            resource_id_pattern: vals[:resource_id_pattern], # Can be +nil+
            asset_type: asset_type,
            collector: collector
          }
        end
      end

      true
    end

    # Parses the JSON object string from the xform artifact row into the following format:
    # [
    #   {
    #     key: String,
    #     value: String
    #   },
    #   ...
    # ]
    def parse_tags(row)
      tags_json = JSON.parse(@stream_adapter.column_value_for_row(row, self.class.required_columns[:tags]))
      out_tag_list = []
      prefix, prefix_length = filter_tag_key_prefix
      tags_json.each do |tag_key, tag_val|
        if prefix.nil?
          out_tag_list << {key: tag_key, value: tag_val}
        elsif tag_key.start_with?(prefix)
          out_tag_list << {key: tag_key[prefix_length..-1], value: tag_val}
        end
      end
      out_tag_list
    end

    # Return the prefix and prefix length to filter tags by. Only tags that pass the filter are added to the job.
    def filter_tag_key_prefix
      return nil, nil
    end

    # Iterates through the job set hash to retrieve all of the fanout jobs that are ready
    # to be yielded
    def get_fanout_jobs!
      fanout_jobs = []
      @job_sets.values.each do |asset_hash|
        asset_hash.values.each do |job_set|
          fanout_jobs.concat(job_set.get_fanout_jobs!(@first_bill_time, @last_bill_time, self.class.cloud_name))
        end
      end
      fanout_jobs
    end

    # Marks the product as completed and then iterates through the job set hash
    # to set the current set of products that have been completed. Job sets will
    # use this information to generate any fanout jobs that can be generated.
    def finish_product
      # Add the last product to the set of processed products to allow should fanout checks to pass
      @products_processed.add(@current_product)

      @job_sets.values.each do |asset_hash|
        asset_hash.values.each do |job_set|
          job_set.finished_product(@products_processed)
        end
      end
    end

    # Checks to ensure that all required columns are present in artifact
    # Logs and raises error if any of the required columns are not present
    def enforce_required_columns(stream_adapter)
      column_check = self.class.required_columns.values.all? do |col|
        stream_adapter.column_exists?(col)
      end

      unless column_check
        message = 'Xform artifact missing required column.'
        logger.error "#{message} Path: #{stream_adapter.file_path}, Customer: #{stream_adapter.customer_id}"
        alert.error(
          message,
          s3_path: stream_adapter.file_path,
          customer_id: stream_adapter.customer_id,
          cloud: self.class.cloud_name
        )
        raise ::Platform::NonRetryableError, orig_exception: message
      end
    end

    # Parses the first and last bill times from the first row of the xform artifact
    # If there is an error parsing the bill times, raise an error
    def parse_bill_times(stream_adapter, row)
      first_bill_time_val = stream_adapter.column_value_for_row(row, self.class.required_columns[:start_time])
      last_bill_time_val = stream_adapter.column_value_for_row(row, self.class.required_columns[:most_recently_found_time])

      begin
        first_bill_time_int = Time.parse(first_bill_time_val).to_i
        last_bill_time_int = Time.parse(last_bill_time_val).to_i
        return first_bill_time_int, last_bill_time_int
      rescue ArgumentError => ex
        message = 'Bill time failed to be converted to a time.'
        logger.error "#{message} Path: #{stream_adapter.file_path}, Customer: #{stream_adapter.customer_id}"
        alert.error(
          message,
          s3_path: stream_adapter.file_path,
          customer_id: stream_adapter.customer_id,
          first_bill_time: first_bill_time_val,
          last_bill_time: last_bill_time_val,
          cloud: self.class.cloud_name
        )
        raise ::Platform::NonRetryableError, orig_exception: ex
      end
    end

    # Check if the resource should be omitted from the jobs
    def ignore_resource?(resource_id, account_id, product_name)
      false
    end

    # Check if the product should be ignored when it comes to determining global last bill time
    def ignore_bill_time?(product_name)
      false
    end

    # Get asset type information for a given resource
    # If no (supported) type matches, return +nil+.
    # Params:
    # - +product_name+:: The +ProductName+ value from the bill. Matched directly.
    # - +res_id+:: The +ResourceId+ value from the bill. Compared against regular expressions for each type.
    # Returns:
    # - A hash of the following structure:
    #   {
    #      resource_id_pattern: regex representing the format of the resource ID Can be +nil+
    #      asset_type: A unique array of [service, asset type],
    #      collector: The registered derived collector class that exposes additional methods with config info
    #   }
    # - +nil+ if the type is not supported.
    def lookup_asset_info(product_name, res_id)
      return nil unless @pnames_to_asset_info.key?(product_name) # Unsupported

      @pnames_to_asset_info[product_name].each do |vals|
        # Product matches and there's no regex, so this is the right asset.
        return vals if vals[:resource_id_pattern].nil?

        # Attempt to match on the regex
        return vals if vals[:resource_id_pattern] =~ res_id
      end

      nil # Unsupported asset type
    end
  end
end
