require 'gdas/assets/aws/ec2/context'

module Gdas::Assets::Aws::Ec2
  module Instances

    class DerivedCollector < ::Gdas::Collection::AwsDerivedCollector
      include Context
      EC2_MIN_ID = 'i-00000000000000000'
      EC2_MAX_ID = 'i-fffffffffffffffff'

      class << self
        def asset
          'instance'
        end

        def gdas_artifact_name
          'gdas_ec2_instances'
        end

        def job_partition_column
          'ResourceId'
        end

        def partition_val_min
          EC2_MIN_ID
        end

        def partition_val_max
          EC2_MAX_ID
        end

        def increment_partition_val(val)
          return EC2_MAX_ID if val == EC2_MAX_ID
          hex_str = val.split('i-')[1]
          right_padded_hex_str = "#{hex_str}#{'0' * (17 - hex_str.size)}"
          "i-%017x" % (right_padded_hex_str.to_i(16) + 1)
        end
      end

      register_derived_collector(
        service: service,
        asset: asset,
        product_name: 'Amazon Elastic Compute Cloud',
        resource_id_pattern: /^i-/,
        format: 1
      )
    end

    class DerivedProcessor < ::Gdas::Processing::Aws::AwsDerivedAssetProcessor
      include ::Gdas::Helpers::Ec2Os
      include ::Gdas::Helpers::Ec2It
      include ::Gdas::Processing::EmojiHelper

      register(
        source: AWS_GDAS_BILL,
        service: DerivedCollector.service,
        asset: DerivedCollector.asset,
        format: 1
      )

      model ::AwsInstance

      asset_identifiers(
        { dso: :resource_id, db: :instance_id }
      )

      def load_gdas_tag_cache
        @tag_cache = {}
        # Read shard
        ::Multidb.use(@customer.read_shard) do
          @ids.each_slice(1000) do |chunk|
            tags = chunked_tag_query(chunk).each_with_object({}) do |t, h|
              h[[t.key, t.value]] = t
            end
            @tag_cache.merge!(tags)
          end
        end
      end

      def chunked_tag_query(id_chunk)
        ::AwsTag.select(%w(aws_tags.id `key` `value` `customer_tag`))
          .joins('INNER JOIN aws_instances_aws_tags ON aws_instances_aws_tags.aws_tag_id = aws_tags.id')
          .where(customer_id: @customer_id, customer_tag: nil)
          .where(aws_instances_aws_tags: {
            aws_instance_id: [id_chunk]
          })
      end

      class << self
        def job_partition_field
          'instance_id'
        end

        def updated_at_date_field
          'updated_at'
        end
      end

      def custom_fields
        {
          name: -> (asset, dso, job_options) { process_instance_name(dso) },
          aws_availability_zone_id: -> (asset, dso, job_options) { process_avail_zone_id(asset, dso, job_options) },
          is_spot: -> (asset, dso, job_options) { process_is_spot(dso) },
          aws_instance_type_id: -> (asset, dso, job_options) { process_instance_type(asset, dso, job_options) },
          tenancy: -> (asset, dso, job_options) { process_tenancy(dso) },
          launch_time: -> (asset, dso, job_options) { process_launch_time(asset, dso) },
          is_ebs_optimized: -> (asset, dso, job_options) { process_is_ebs_optimized(dso) },
          state: -> (asset, dso, job_options) { process_state(asset) }
        }
      end

      def process_instance_name(dso)
        name = remove_emoji_characters(dso[:custom_columns]['InstanceName'])
        return dso[:resource_id] if name.blank?
        name
      end

      def process_is_spot(dso)
        dso[:custom_columns]['IsSpot'] == 'true'
      end

      def process_avail_zone_id(asset, dso, job_options)
        az_name = dso[:custom_columns]['AvailabilityZone']
        return nil if self.process_is_spot(dso) || az_name.blank?

        avail_zone = common_assets.zones_by_name[az_name]

        if avail_zone.nil?
          logger.warn("Availability zone not found: #{az_name}, skipping derivation of EC2 instance: #{dso[:resource_id]}")
          statsd.increment 'gdas.assets.new_az', tags: %W(az:#{az_name}) unless asset.persisted?

          customer_id = job_options.dig(:customer_id)
          rollbar_with_redis_set(
            "#{customer_id}:gdas:ec2:new_az",
            az_name,
            'G-DAS failed to derive an EC2 instance with a missing AZ.',
            customer_id: customer_id,
            aws_account_id: job_options.dig(:aws_account_id),
            instance_id: dso[:resource_id],
            availability_zone: az_name
          )

          # Do not process EC2 instances that are not spot and do not have an availability zone
          throw :skip
        end

        avail_zone[:id]
      end

      def process_instance_type(asset, dso, job_options)
        custom_cols = dso[:custom_columns]
        usage_type = custom_cols['InstanceTypeUsageType']
        operation = custom_cols['InstanceTypeOperation']
        description = custom_cols['InstanceTypeDescription']
        tenancy = custom_cols['Tenancy']
        region_name = dso[:region]

        if operation.blank? || (usage_type.blank? && description.blank?)
          logger.debug "Not enough information to derive instance type for #{dso[:resource_id]}"
          throw :skip_field_update if asset.persisted?
          throw :skip
        end

        instance_type = lookup_instance_type(usage_type, operation, description, region_name, tenancy)

        if instance_type.nil?
          instance_type_name = get_instance_type_from_ut_and_desc(usage_type, description)

          tags = %W(
            asset_type:ec2_instance
            derived:true
            region_name:#{region_name || 'us-east-1'}
            api_name:#{instance_type_name}
            os:#{get_os_by_operation(operation)}
            tenancy:#{tenancy}
          )
          statsd.increment('missing_aws_instance_type', tags: tags)
          logger.warn do
            "Unsupported instance type. Customer: #{job_options.dig(:customer_id)}, "\
            "account: #{job_options.dig(:aws_account_id)}, "\
            "resource ID: #{dso[:resource_id]}, operation: #{operation}, details: #{tags}"
          end

          # Do not update existing instance instance type field to null value
          # https://cloudhealthtech.atlassian.net/browse/AWS-9422
          throw :skip_field_update if asset.persisted?

          # Do not derive instances with unsupported instance type
          throw :skip
        end

        instance_type[:id]
      end

      def process_launch_time(asset, dso)
        launch_time = dso[:start_time]
        return launch_time if asset[:launch_time].nil? || launch_time < asset[:launch_time]
        throw :skip_field_update
      end

      def process_tenancy(dso)
        dso[:custom_columns]['Tenancy']
      end

      def process_is_ebs_optimized(dso)
        dso[:custom_columns]['EbsOptimized'] == 'true'
      end

      def process_state(asset)
        # The inactivated_at field gets updated by the derived processor before the custom fields get updated.
        asset.inactivated_at.nil? ? 'running' : 'terminated'
      end

      def lookup_instance_type(usage_type, operation, description, region_name, tenancy)
        it_name = get_instance_type_from_ut_and_desc(usage_type, description)
        os = get_os_by_operation(operation)
        region = region_name || 'us-east-1'
        common_assets.instance_type_by_key(region, it_name, os, tenancy)
      end

      # Rollbar only if the value provided does not exist in the set under the given redis key. Add value to set with
      # default TTL of 1 day.
      def rollbar_with_redis_set(key, value, message, rollbar_tags)
        unless redis.smembers(key).include?(value)
          alert.error(
            message,
            rollbar_tags
          )
          redis.add_to_set(key, value)
        end
      end
    end
  end
end

