require 'gdas/assets/aws/rds/context'

module Gdas::Assets::Aws::RDS
  module Instances

    class DerivedCollector < ::Gdas::Collection::AwsDerivedCollector
      include Context

      class << self
        def asset
          'instance'
        end

        def gdas_artifact_name
          'gdas_rds_instances'
        end
      end

      register_derived_collector(
        service: service,
        asset: asset,
        product_name: ['Amazon Relational Database Service', 'Amazon RDS Service'],
        resource_id_pattern: /^arn:aws[\w-]*:rds:[^:]*:[^:]*:db:.*/,
        format: 1
      )
    end

    class DerivedProcessor < ::Gdas::Processing::Aws::AwsDerivedAssetProcessor
      include ::Gdas::Helpers::Ec2It
      include ::Gdas::Helpers::RdsUtils

      register(
        source: AWS_GDAS_BILL,
        service: DerivedCollector.service,
        asset: DerivedCollector.asset,
        format: 1
      )

      model ::AwsRdsInstance

      asset_identifiers(
        :aws_region_id,
        :aws_account_id,
        :instance_id
      )

      set_callback :process_assets, :before, :set_identifiers_in_payload

      # arn:aws:rds:us-east-1:844584796563:db:mj-aurora-mysql-oneaz -> mj-aurora-mysql-oneaz
      # Gov cloud is arn:aws-us-gov:rds...
      RDS_INSTANCE_ID_FROM_RESOURCE_ID_PATTERN = /arn:aws[\w-]*:rds:[^:]*:[^:]*:db:(.*)/

      class << self
        def updated_at_date_field
          'updated_at'
        end

        def resource_id_sha1_source_field
          :instance_arn
        end
      end

      # Must populate the dso with instance_id and aws_account_id so that the asset_identifiers work as expected, e.g.
      # for cache building, asset lookup, etc.
      def set_identifiers_in_payload
        data.reject! do |dso|
          instance_id = extract_instance_id(dso[:resource_id], dso[:aws_region_id])
          if instance_id && account
            dso[:instance_id] = instance_id
            dso[:aws_account_id] = account.id
            false # Good values. Keep in +data+ for processing.
          else
            true
          end
        end
      end

      def base_asset_queries
        raise unless @first_bill_time # Just in case. This is built and checked in +extract_payload_times+.

        lower_bound, upper_bound = get_partition_bounds

        partition_lower_clause = "#{self.class.job_partition_field} >= '#{lower_bound}'"
        partition_upper_clause = "#{self.class.job_partition_field} <= '#{upper_bound}'"
        partition_not_null_clause = "#{self.class.job_partition_field} IS NOT NULL"

        # The aws_rds_instances table can contain entries with the same region, account, and instance_id, in order to
        # select the correct one, we add an ordering clause such that when the cache is built, the entry for each
        # region, account, and instance_id key will be the one with the highest "priority".
        #
        # Priority hierarchy is active first, then collected, then most recent created_at_rds time
        [
          self.class.klass_model
              .where(@lookup_attributes)
              .active
              .where(partition_not_null_clause)
              .where(partition_lower_clause)
              .where(partition_upper_clause)
              .order('is_derived DESC, created_at_rds')
        ]
      end

      # Override lookup method to add the ordering clause to prioritize instances in the same way as the base_asset_query
      def lookup_asset(dso)
        identifier, bad_prop = dso_to_identifier dso, self.class.model_identifiers
        asset = @assets[identifier]
        return asset if asset

        attrs = dso_to_attrs dso, self.class.model_identifiers
        asset = self.class.klass_model
                    .unscoped
                    .where(@lookup_attributes)
                    .where(attrs)
                    .order('is_active DESC, is_derived, created_at_rds DESC')
                    .first
        return asset if asset

        # if we have at least one property with a bad value, call skip.
        # we already issued a rollbar in dso_to_identifier
        throw :skip unless bad_prop.nil?

        # TODO: first_or_initialize broken in ruby 2.3.0
        # https://github.com/rails/protected_attributes/issues/53#issuecomment-174062735

        # Falling back to creating a new asset.
        self.class.klass_model.new.tap do |t|
          [@lookup_attributes, attrs].each do |y|
            y.each_pair do |k, v|
              t.send :"#{k}=", v
            end
          end
          t.is_derived = true
        end
      end

      # Overriding this since not all RDS instances in the job partition range are loaded into the cache due to
      # duplicate entries with the same region, account, and instance_id values. This query ensures that we are
      # cleaning up the old duplicate derived instances.
      def assets_for_termination
        lower_bound, upper_bound = get_partition_bounds
        partition_lower_clause = "#{self.class.job_partition_field} >= '#{lower_bound}'"
        partition_upper_clause = "#{self.class.job_partition_field} <= '#{upper_bound}'"
        partition_not_null_clause = "#{self.class.job_partition_field} IS NOT NULL"

        self.class.klass_model
            .where(@lookup_attributes)
            .where(is_active: true, is_derived: true)
            .where("updated_at <= ?", @last_bill_time - self.class.gap_for_declaring_inactive.hours)
            .where(partition_not_null_clause)
            .where(partition_lower_clause)
            .where(partition_upper_clause)
      end

      # Overriding just to get around the check that the asset existed in the cache before inactivating it.
      def inactivate_terminated_assets
        terminate_assets = assets_for_termination
        return if terminate_assets.blank?

        logger.debug "Inactivating #{terminate_assets.length} assets..."
        terminate_assets.each do |asset|
          # This keeps the +updated_at+ timestamp from changing when the asset is saved. For derived assets,
          # +updated_at+ means the last time the asset was seen in the bill, *not* the last time a change for
          # the asset was persisted to the database.
          asset.updated_at_will_change!
          ::CloudHealth::Processing::Actions::TerminateAsset.call asset: asset,
                                                                  inactivated_at: asset.updated_at,
                                                                  audit_log: audit_log,
                                                                  account: account
        end
      end

      def custom_fields
        {
          instance_arn: -> (asset, dso, job_options) { process_instance_arn(asset, dso) },
          flavor_id: -> (asset, dso, job_options) { process_flavor_id(dso) },
          multi_az: -> (asset, dso, job_options) { process_multi_az(dso) },
          engine: -> (asset, dso, job_options) { process_engine(asset, dso) },
          created_at_rds: -> (asset, dso, job_options) { process_created_at_rds(asset, dso) },
          license_model: -> (asset, dso, job_options) { process_license_model(asset, dso) },
          aws_instance_type_id: -> (asset, dso, job_options) { process_instance_type_id(asset, dso, job_options) },
          storage_type: -> (asset, dso, job_options) { process_storage_type(asset, dso, job_options) },
          state: -> (asset, dso, job_options) { process_state(asset, dso) },
        }
      end

      def forced_custom_fields
        {
          multi_az: -> (asset, dso, job_options) { process_forced_multi_az(dso) }
        }
      end

      def process_instance_arn(asset, dso)
        throw :skip_field_update if asset.instance_arn.present?
        dso[:resource_id]
      end

      def extract_instance_id(resource_id, region_id)
        matched_instance_id = RDS_INSTANCE_ID_FROM_RESOURCE_ID_PATTERN.match(resource_id)
        return nil if matched_instance_id.nil?
        instance_id = matched_instance_id[1]

        instance_id = instance_id.rpartition(/-[\w]+-[\w]+-[\w]+/)[0] if has_az_suffix?(instance_id, region_id)
        instance_id
      end

      def process_storage_type(asset, dso, job_options)
        throw :skip_field_update if asset[:storage_type] == 'aurora'

        engine = engine_from_operation(dso[:custom_columns]['InstanceTypeOperation'])
        return 'aurora' if %w(aurora aurora-postgresql).include?(engine)

        usage_type = dso[:custom_columns]['StorageUsageType']
        storage_type = storage_type_from_usage_type(usage_type)

        if storage_type.nil?
          cache_key = "gdas:rds:storage_usage_type:#{usage_type}"
          unless redis.get(cache_key)
            alert.error(
              'G-DAS found unexpected RDS storage record.',
              customer_id: job_options.dig(:customer_id),
              account_id: job_options.dig(:aws_account_id),
              usage_type: usage_type
            )
            redis.set(cache_key, true)
          end
        end

        storage_type
      end

      def process_engine(asset, dso)
        throw :skip_field_update unless asset[:engine].nil?
        engine_from_operation(dso[:custom_columns]['InstanceTypeOperation'])
      end

      def process_license_model(asset, dso)
        throw :skip_field_update unless asset[:license_model].nil?
        license_from_operation(dso[:custom_columns]['InstanceTypeOperation'])
      end

      def process_multi_az(dso)
        dso[:custom_columns]['MultiAz'] == 'true'
      end

      def process_forced_multi_az(dso)
        return true if has_az_suffix?(dso[:resource_id], dso[:aws_region_id])
        throw :skip_field_update
      end

      def process_created_at_rds(asset, dso)
        launch_time = dso[:start_time]
        return launch_time if asset[:created_at_rds].nil? || launch_time < asset[:created_at_rds]
        throw :skip_field_update
      end

      def process_flavor_id(dso)
        cols = dso[:custom_columns]
        instance_type_for_rds_sf(cols['InstanceTypeUsageType'], cols['InstanceTypeDescription'])
      end

      def process_instance_type_id(asset, dso, job_options)
        cols = dso[:custom_columns]
        operation = cols['InstanceTypeOperation']
        api_name = process_flavor_id(dso)
        region_name = dso[:region]
        multi_az = process_multi_az(dso)

        engine_license = map_operation_to_engine(operation)

        if engine_license.nil?
          alert.error(
            'G-DAS found unexpected engine or license for RDS instance during instance type derivation.',
            customer_id: job_options.dig(:customer_id),
            operation: operation
          )
          throw :skip_field_update if asset.persisted?
          throw :skip
        end

        license_model = engine_license[:license] == 'none' ? '' : "(#{engine_license[:license]})"
        profile_os = ::AwsPricing::DatabaseType.db_mapping(engine_license[:engine] + license_model, multi_az)

        instance_type = common_assets.instance_type_by_key(region_name, api_name, profile_os, 'default')

        if instance_type.nil?
          tags = %W(
            asset_type:rds_instance
            derived:true
            region_name:#{region_name}
            api_name:#{api_name}
            os:#{profile_os&.tr(' ()', '_')}
            tenancy:default
          )
          statsd.increment('missing_aws_instance_type', tags: tags)
          logger.warn do
            "Unsupported instance type. Customer: #{job_options.dig(:customer_id)}, "\
            "account: #{job_options.dig(:aws_account_id)}, "\
            "resource ID: #{dso[:resource_id]}, engine: #{engine_license[:engine]}, lic_mod: #{license_model}, "\
            "details: #{tags}"
          end

          throw :skip_field_update if asset.persisted?
          throw :skip
        end

        instance_type[:id]
      end

      def process_state(asset, dso)
        # The inactivated_at field gets updated by the derived processor before the custom fields get updated.
        return 'stopped' if asset.inactivated_at.present?
        dso[:most_recently_found_time] == dso[:custom_columns]['MostRecentParsableIt'] ? 'available' : 'stopped'
      end

      DB_M1_SMALL = 'db.m1.small'.freeze
      # Returns a String for an instance's instance type. By default, we return
      # m1.small
      def instance_type_for_rds_sf(usage_type, item_desc)
        instance_type =
          if item_desc =~ /UsageType:/
            item_desc =~ /UsageType: .*Usage:([^ ]*)/
            $1 || DB_M1_SMALL
          else
            usage_type =~ /Usage:([^ ]*)/
            $1 || DB_M1_SMALL
          end

        if match = instance_type.match(/(^.*\.[0-9]*[x]?l)([\.\z].*$)/)
          instance_type = match.captures[0] + "arge" + match.captures[1]
        end

        instance_type
      end

      def engine_from_operation(operation)
        engine_license = map_operation_to_engine(operation)
        return if engine_license.nil?

        engine = engine_license[:engine]
        engine == POSTGRESQL ? 'postgres' : engine
      end

      def has_az_suffix?(resource_id, region)
        region = common_assets.region(region)
        az_names = region.aws_availability_zones.map { |az| az.name }
        resource_id.ends_with?(*az_names)
      end

      def license_from_operation(operation)
        engine_license = map_operation_to_engine(operation)
        return if engine_license.nil?

        return 'postgresql-license' if [POSTGRESQL, AURORA_POSTGRESQL].include?(engine_license[:engine])

        case engine_license[:license]
        when 'byol'
          'bring-your-own-license'
        when 'li'
          'license-included'
        else
          'general-public-license'
        end
      end

      def storage_type_from_usage_type(usage_type)
        case usage_type
        when /:StorageUsage$/, /:Multi-AZ-StorageUsage$/, /:Mirror-StorageUsage$/
          'standard'
        when /:PIOPS-Storage$/, /:Multi-AZ-PIOPS-Storage$/, /:Mirror-PIOPS-Storage$/
          'io1'
        when /:GP2-Storage$/, /:Multi-AZ-GP2-Storage$/, /:Mirror-GP2-Storage$/
          'gp2'
        else
          nil # Don't guess.
        end
      end
    end
  end
end
