
module Gdas::Assets::GenericDerivedAssetScanner::CsvStream

  # Interface for the Xform artifact streamers. Also contains a factory method to get the appropriate streamer, based on
  # if the given filepath is local or on S3.
  class CsvStreamer
    include CsvAnalyzerUtilities

    attr_reader :file_path, :customer_id

    # Factory method to create a +CsvStreamer+.
    #
    # Params:
    # - +customer_id+:: FixNum. Use to add information to Rollbars.
    # - +file_path+:: String. Path to the zipped CSV file, on S3 or local machine.
    # - +access_key+:: String. AWS access key used to retrieve the file from S3.
    # - +secret_key+:: String. AWS secret key used to retrieve the file from S3.
    # - +session_token+:: (Optional) String. AWS session token used to retrieve the file from S3. If included,
    #                       it's passed through to the +EmbeddedStreamFilter+. Some accounts do not use
    #                       session tokens; CHT does.
    # - +pattern+:: (Optional) String. Regular expression applied to the "ProductName" column to filter
    #                 DBR lines. If left as an empty string, no filtering is applied.
    # - +metric_file_type+:: (Optional) String. Identifier to use when posting metrics about file streaming.
    # - +**options+:: Hash. Needed for Dry RB to apply the additional +Import+s above.
    #
    # Returns:
    # - +LocalCsvStreamer+ if given a non-S3 path and running in sandbox
    # - +S3StreamFilterAdapter+ otherwise
    def self.get_csv_streamer(customer_id, file_path, access_key, secret_key, session_token: nil, pattern: '', metric_file_type: nil, **options)
      is_sandbox = ImportRegistry.resolve('runtime').sandbox?
      if !is_sandbox && !file_path.start_with?('s3:')
        raise 'Given non-S3 xform artifact path on non-sandbox environment'
      elsif is_sandbox && !file_path.start_with?('s3:')
        LocalCsvStreamer.new(customer_id, file_path)
      else
        S3StreamFilterAdapter.new(
          customer_id,
          file_path,
          access_key,
          secret_key,
          session_token: session_token,
          pattern: pattern,
          metric_file_type: metric_file_type
        )
      end
    end

    # Start the stream and (optionally) parse the header row.
    # This *must* be called before using the stream with +data_rows_each+.
    #
    # Params:
    # - +process_header_row+:: Bool. Default +true+. If +false+, skip the processing of the header row.
    #                            With header processing disabled, the column helpers below will raise errors.
    #
    # Returns:
    # - +true+ if the stream is started.
    #
    # Raises:
    # - +::Platform::NonRetryableError+ if the file is empty (also logged/alerted)
    def start_stream(process_header_row: true)
      raise NotImplementedError
    end

    # Call the provided block on each row of the CSV file. If header processing is enabled (default),
    # this will only run on the data rows. If header processing is disabled, the header row will be
    # returned on first run.
    #
    # +start_stream+ *must* be called before this.
    #
    # Yields:
    # - List of strings representing the data in the row.
    #
    # Examples:
    #   adapter.data_rows_each do |row|
    #     # Process row
    #     value = adapter.column_value_for_row(row, 'Column Name')
    #     # ...
    #   end
    def data_rows_each
      raise NotImplementedError
    end

    # If header processing is enabled, retrieve a list of all column headers in the CSV.
    # A header is simply the first line of the CSV.
    #
    # Returns:
    # - List of strings, where each string is the header of a row. Order is *not* guaranteed.
    def column_headers
      raise 'Header not processed during start_stream.' unless @column_index
      @column_index.keys
    end

    # If header processing is enabled, check if the given column header exists in the CSV file.
    #
    # Params:
    # - +column_name+:: String. Header of the column to verify.
    #
    # Returns:
    # - +true+ if the column name is in the set of processed headers.
    # - +false+ otherwise.
    def column_exists?(column_name)
      raise 'Header not processed during start_stream.' unless @column_index
      @column_index.key?(column_name)
    end

    # If header processing is enabled, get the value of the given row for the given column.
    #
    # Params:
    # - +row+:: List of strings. The row to extract the value from.
    # - +column_name+:: String. Header of the column to find.
    #
    # Returns:
    # - String value of the row in the given column.
    def column_value_for_row(row, column_name)
      raise 'Header not processed during start_stream.' unless @column_index
      raise 'Column name not in file.' unless @column_index.key?(column_name)

      row[@column_index[column_name]]
    end

  end
end
