require 'zip'
require 'csv'

module Gdas::Assets::GenericDerivedAssetScanner::CsvStream

  # An alternative to the +S3StreamFilterAdapter+ for use in local development: this allows reading from a local
  # xform artifact instead of from S3.
  class LocalCsvStreamer < CsvStreamer
    include Import['logger']

    # Create a new CSV streamer.
    #
    # Params:
    # - +customer_id+:: FixNum. Use to add information to Rollbars.
    # - +file_path+:: String. Path to the zipped CSV file on local machine.
    # - +**options+:: Hash. Needed for Dry RB to apply the additional +Import+s above.
    def initialize(customer_id, file_path, **options)
      super(options) # For Dry RB initialization
      @file_path = file_path
      @customer_id = customer_id
    end

    # Start the local stream and (optionally) parse the header row.
    # This *must* be called before using the stream with +data_rows_each+.
    #
    # Params:
    # - +process_header_row+:: Bool. Default +true+. If +false+, skip the processing of the header row.
    #                            With header processing disabled, the column helpers below will raise errors.
    #
    # Returns:
    # - +true+ if the stream is started.
    #
    # Raises:
    # - +::Platform::NonRetryableError+ if the file is empty (also logged/alerted)
    def start_stream(process_header_row: true)
      raise 'Stream already started.' if @csv_stream

      begin
        @file = get_file
        @file_stream = @file.glob('*.csv').first.get_input_stream
        @csv_stream = CSV.new(@file_stream)
      rescue ::Zip::Error
        message = 'Error opening CSV file.'
        logger.error "#{message} Path: #{@file_path}, Customer: #{@customer_id}"
        raise ::Platform::NonRetryableError, orig_exception: message
      end


      if process_header_row
        # Get the first row and parse the headers from it.
        # Build a column label index lookup based on this row.
        header_row = @csv_stream.shift
        unless header_row
          message = 'Attempted to stream empty CSV file.'
          logger.error "#{message} Path: #{@file_path}, Customer: #{@customer_id}"
          raise ::Platform::NonRetryableError, orig_exception: message
        end

        @column_index = build_column_label_index(header_row)
      end

      true
    end

    def get_file
      Zip::File.open(@file_path)
    end

    # Call the provided block on each row of the CSV file. If header processing is enabled (default),
    # this will only run on the data rows. If header processing is disabled, the header row will be
    # returned on first run.
    #
    # +start_stream+ *must* be called before this.
    #
    # Yields:
    # - List of strings representing the data in the row.
    #
    # Examples:
    #   adapter.data_rows_each do |row|
    #     # Process row
    #     value = adapter.column_value_for_row(row, 'Column Name')
    #     # ...
    #   end
    def data_rows_each
      raise 'Stream not started.' unless @csv_stream
      raise 'No block provided.' unless block_given?
      line_count = 0

      until @csv_stream.eof?
        yield(@csv_stream.shift)
        line_count += 1
        logger.warn("Processed #{line_count} lines (#{row_timing.rate}/s) from local file: '#{@file_path}'.") if (line_count % 1_000_000).zero?
      end
      @csv_stream.close
      @file_stream.close
      @file.close

      logger.warn("Done with all (#{line_count}) rows for local file: '#{@file_path}'.")
    end
  end
end
