# frozen_string_literal: true

require 'hitimes'

module Gdas::Assets::GenericDerivedAssetScanner::CsvStream
  # S3 Stream Filter Adapter: A wrapper around the +EmbeddedStreamFilter+ Java class used to help the
  # CSV analyzers stream, filter, and process CSV files.
  # This class provides a convenient interface to enumerate the rows of a CSV file, parse the headers,
  # and lookup the values at a given column. It also allows the +Analyzer+s to avoid managing credentials and
  # streaming logic.
  #
  # Examples:
  #   adapter = S3StreamFilterAdapter.new(creds, path)
  #   adapter.start_stream # Parses the headers to create a lookup
  #   adapter.data_rows_each do |row|
  #     # Process row
  #     value = adapter.column_value_for_row(row, 'Column Name')
  #   end
  class S3StreamFilterAdapter < CsvStreamer
    include Import['logger', 'alert', 'statsd', 'gdas.s3_csv_stream_filterer']

    # Create a new stream adapter.
    #
    # Params:
    # - +customer_id+:: FixNum. Use to add information to Rollbars.
    # - +file_path+:: String. Path to the zipped CSV file in S3.
    #                 Format: "s3://cloudhealth-detailed-bills//2017/33...11/84458479...2018-05.csv.zip"
    # - +access_key+:: String. AWS access key used to retrieve the file from S3.
    # - +secret_key+:: String. AWS secret key used to retrieve the file from S3.
    # - +session_token+:: (Optional) String. AWS session token used to retrieve the file from S3. If included,
    #                       it's passed through to the +EmbeddedStreamFilter+. Some accounts do not use
    #                       session tokens; CHT does.
    # - +pattern+:: (Optional) String. Regular expression applied to the "ProductName" column to filter
    #                 DBR lines. If left as an empty string, no filtering is applied.
    # - +metric_file_type+:: (Optional) String. Identifier to use when posting metrics about file streaming.
    # - +**options+:: Hash. Needed for Dry RB to apply the additional +Import+s above.
    def initialize(customer_id, file_path, access_key, secret_key, session_token: nil, pattern: '', metric_file_type: nil, **options)
      super(options) # For Dry RB initialization
      raise 'No credentials provided.' if access_key.blank? || secret_key.blank?
      raise 'No S3 path provided.' if file_path.blank?

      @credential_string = create_credential_string(access_key, secret_key, session_token)
      @file_path = file_path
      @pattern = pattern
      @customer_id = customer_id
      @metric_file_type = metric_file_type
    end

    # Start the stream from S3 and (optionally) parse the header row.
    # This *must* be called before using the stream with +data_rows_each+.
    # Note that this call will open a network connection to S3 that will remain open until all rows have
    # been consumed. As such, call +start_stream+ just before using the headers and data of the streamed
    # file instead of starting the stream early.
    #
    # Params:
    # - +process_header_row+:: Bool. Default +true+. If +false+, skip the processing of the header row.
    #                            With header processing disabled, the column helpers below will raise errors.
    #
    # Returns:
    # - +true+ if the stream is started.
    #
    # Raises:
    # - +::Platform::NonRetryableError+ if the file is empty (also logged/alerted)
    def start_stream(process_header_row: true)
      raise 'Stream already started.' if @filter_stream

      begin
        @filter_stream = s3_csv_stream_filterer.new(
          @credential_string,
          @file_path,
          @pattern
        )
      rescue ::Java::ComAmazonawsServicesS3Model::AmazonS3Exception => ex
        # S3 path does not exist, for example.
        # For now, just Rollbar this until we know what possible error conditions we should handle.
        message = 'S3 exception when trying to start stream.'
        logger.error "#{message} Path: #{@file_path}, Customer: #{@customer_id}; Ex: #{ex.class}: #{ex.message}"
        alert.exception ex, s3_path: @file_path, customer_id: @customer_id
        raise ::Platform::NonRetryableError, orig_exception: ex
      end

      if process_header_row
        # Get the first row and parse the headers from it.
        # Build a column label index lookup based on this row.
        header_row = @filter_stream.next_row
        unless header_row
          message = 'Attempted to stream empty CSV file.'
          logger.error "#{message} Path: #{@file_path}, Customer: #{@customer_id}"
          alert.error message, s3_path: @file_path, customer_id: @customer_id
          raise ::Platform::NonRetryableError, orig_exception: message
        end

        @column_index = build_column_label_index(header_row)
      end

      true
    end

    # Call the provided block on each row of the CSV file. If header processing is enabled (default),
    # this will only run on the data rows. If header processing is disabled, the header row will be
    # returned on first run.
    #
    # +start_stream+ *must* be called before this.
    #
    # Yields:
    # - List of strings representing the data in the row.
    #
    # Examples:
    #   adapter.data_rows_each do |row|
    #     # Process row
    #     value = adapter.column_value_for_row(row, 'Column Name')
    #     # ...
    #   end
    def data_rows_each
      raise 'Stream not started.' unless @filter_stream
      raise 'No block provided.' unless block_given?
      line_count = 0
      row_timing = Hitimes::TimedMetric.new('row processing time')

      until (row = @filter_stream.next_row).nil?
        row_timing.start
        yield(row.to_a)
        row_timing.stop

        line_count += 1
        logger.warn("Processed #{line_count} lines (#{row_timing.rate}/s) from S3 path: '#{@file_path}'.") if (line_count % 1_000_000).zero?
      end

      logger.warn(
        "Done with all (#{line_count}) rows for S3 path: '#{@file_path}'. Stats (not including network time): "\
        "mean: #{row_timing.mean}, min: #{row_timing.min}, max: #{row_timing.max}, rate: #{row_timing.rate}"
      )

      if @metric_file_type
        statsd.increment(
          's3_stream_filter.lines_processed',
          by: line_count,
          tags: ["file_type:#{@metric_file_type}"]
        )
      end
    end

    private

    # Build the passthrough credential string expected by the +EmbeddedStreamFilter+.
    #
    # Params:
    # - +access_key+:: String. AWS access key.
    # - +secret_key+:: String. AWS secret key.
    # - +session_token+:: (Optional) String. AWS session token. If included, it's passed through to the
    #                       +EmbeddedStreamFilter+.
    #
    # Returns:
    # - String. Credential string expected by the +EmbeddedStreamFilter+ to retrieve the file from S3.
    #     Format: "passthrough:#{access_key}:#{secret_key}:#{session_token}"
    def create_credential_string(access_key, secret_key, session_token)
      if session_token
        "passthrough:#{access_key}:#{secret_key}:#{session_token}"
      else
        "passthrough:#{access_key}:#{secret_key}"
      end
    end
  end
end
