# frozen_string_literal: true

require 'gdas/assets/generic_derived_asset_scanner/context'
require 'gdas/artifact_analyzers/xform_analyzer'
require 'gdas/clouds/cloud_config'
require 'time'

module Gdas::Assets::GenericDerivedAssetScanner
  module Scanner
    class Collector < ::Gdas::Collection::Executor
      # Customer context sets the DB shard and leases a connection for the whole job.
      # We can probably do better than that, but works for now.
      include ::Tasks::Context::Customer
      include Gdas::ArtifactAnalyzers::RegisteredAnalyzers
      include Gdas::Processing::AssetCaches

      include Context
      include Import['gdas.iam_credentials', 'gdas.s3_csv_stream_filterer']

      # +invoke+ is the highest level available. It is defined in +::Tasks::Executor+.
      set_callback :invoke, :before, -> { log_job_details('Starting') }
      set_callback :invoke, :after, -> { log_job_details('Stopping') }

      class << self
        def asset
          'generic_derived_asset_scanner'
        end
      end

      def log_job_details(action)
        logger.warn do
          "#{action} G-DAS collection job for customer: #{@customer_id}, "\
          "from xform path: #{@options.dig(:gdas_xform_s3_path)}"
        end
      end

      def execute
        # This path comes from the Kafka message from the superxform query.
        gdas_xform_s3_path = @options[:gdas_xform_s3_path]

        if gdas_xform_s3_path.blank?
          message = 'Required S3 path missing from context.'
          logger.error "#{message} gdas_xform_s3_path: '#{gdas_xform_s3_path}', Customer: #{@customer_id}"
          alert.error message, gdas_xform_s3_path: gdas_xform_s3_path, customer_id: @customer_id
          raise ::Platform::NonRetryableError, orig_exception: message
        end

        artifact_name = get_artifact_name(gdas_xform_s3_path)
        analyzer_class = ARTIFACT_TO_ANALYZER[artifact_name]
        if analyzer_class.nil?
          message = 'Unregistered G-DAS xform artifact, skipping.'
          logger.warn "#{message} gdas_xform_s3_path: '#{gdas_xform_s3_path}', Customer: #{@customer_id}"
          alert.warn message, gdas_xform_s3_path: gdas_xform_s3_path, customer_id: @customer_id
          raise ::Platform::NonRetryableError, orig_exception: message
        end

        # Build the (product name) whitelist based on assets registered for this artifact type
        product_names = []
        ::Gdas::Assets::GenericDerivedAssetScanner::Registry.instance.each do |vals, collector|
          if collector.gdas_artifact_name == artifact_name
            vals[:product_name].each do |product_name|
              product_names << Regexp.escape(product_name)
            end
            # Format pattern as ".*<ProductName>.*" for `product_match_name`
            # Note: `EmbeddedStreamFilter` does exact match to `filter_whitelist_pattern` and product name
            product_names << vals[:product_match_name].source if vals[:product_match_name]
          end
        end
        filter_whitelist_pattern = product_names.join('|')

        # For development (when adding a new asset, for example), it can be much faster to override this
        # whitelist with only the relevant product name. This will make the collection stage significantly
        # faster and cause fewer processing jobs to be created.
        # For example:
        # filter_whitelist_pattern = 'Amazon AppStream'

        # If no asset types are registered, don't bother continuing.
        if filter_whitelist_pattern.empty?
          message = 'No GDAS asset types are registered.'
          logger.error "#{message} Customer: #{@customer_id}"
          alert.error message, customer_id: @customer_id
          raise ::Platform::NonRetryableError, orig_exception: message
        end

        logger.debug("ProductName whitelist: #{filter_whitelist_pattern}")

        # Ready the CHT credentials
        iam_credentials.refresh_credentials

        stream_filter_adapter = Gdas::Assets::GenericDerivedAssetScanner::CsvStream::CsvStreamer.get_csv_streamer(
            @customer_id,
            gdas_xform_s3_path,
            iam_credentials.access_key,
            iam_credentials.secret_key,
            session_token: iam_credentials.session_token,
            pattern: filter_whitelist_pattern,
            metric_file_type: artifact_name
        )

        # XformAnalyzer generates fanout jobs with the following structure:
        # {
        #   cloud: ...
        #   account_id: ...
        #   asset_type: ...
        #   first_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
        #   last_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
        #   partition_lower_bound: String,
        #   partition_upper_bound: String,
        #   resources: [
        #     {
        #       resource_id: String,
        #       cloud_account_id: String,
        #       region: String,
        #       asset_type: String,
        #       start_time: DateTime,
        #       most_recently_found_time: DateTime
        #       tags: {tag_key: tag_value},
        #       custom_columns: {"ColumnName" => column_val}
        #     },
        #     ...
        #   ]
        # }
        jobs_processed = 0
        analyzer = analyzer_class.new(stream_filter_adapter, artifact_name)
        analyzer.each_fanout_job do |fanout_job|
          # Call the registered processors for each supported asset type found in the bill.
          fan_out_job_to_processing(fanout_job)
          jobs_processed += 1
        end

        # If no resources were identified, don't bother continuing.
        if jobs_processed == 0
          logger.info("No resources identified in bill. Xform Path: #{gdas_xform_s3_path}, customer: #{@customer_id}")
          raise ::Platform::NonRetryableError
        end

        # Create or update a +JobCompletion+ record, showing the most recent collection time per customer.
        update_job_completion_time

        nil # Jobs have already been sent out, so no need to return a payload.
      end

      private

      def cloud_config
        Gdas::Clouds::CloudConfig.cloud_config(@options[:cloud])
      end

      def get_artifact_name(artifact_path)
        artifact_path.split('/')[-1].sub('.csv.zip', '')
      end

      def update_job_completion_time
        Multidb.use(:master) do
          job = JobCompletion.where(
              job_name: asset,
              job_key: @customer_id.to_s
          ).first_or_initialize
          job.completion_time = DateTime.now
          job.save!
        end
      end

      # Fan out the job returned by the xform analyzer. This works by confirming that the
      # account and asset are valid, and sending out the processing payloads. Accepts the following
      # job format:
      # {
      #   account_id: ...
      #   asset_type: [asset, service], # Two element array containing asset and service
      #   first_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
      #   last_bill_time: FixNum, # Parsed from the time string. Recover with +Time.at+.
      #   partition_lower_bound: String,
      #   partition_upper_bound: String,
      #   resources: [
      #     {
      #       resource_id: String,
      #       cloud_account_id: String,
      #       region: String,
      #       asset_type: String,
      #       start_time: DateTime,
      #       most_recently_found_time: DateTime
      #       tags: {tag_key: tag_value},
      #       custom_columns: {"ColumnName" => column_val}
      #     },
      #     ...
      #   ]
      # }
      def fan_out_job_to_processing(fanout_job)
        collection_job_finish_time = Time.now.to_i

        payload_account_id = self.send(cloud_config.account_id_lookup_func)[fanout_job[:account_id]]

        # If we don't know about this account yet, skip it. Will be derived in the EC2 scanner run daily.
        return unless payload_account_id

        asset_type = fanout_job[:asset_type]

        format = format_for_service_and_asset(@options[:cloud], asset_type)
        if format.nil?
          alert.error(
              'Derived collector for asset is misconfigured.',
              cloud: @options[:cloud],
              service_asset_type: asset_type,
              customer_id: @customer_id
          )
          return
        end

        payload = {
            data_source: data_source_by_cloud(@options[:cloud]),
            cloud: @options[:cloud],
            service: asset_type[0],
            asset: asset_type[1],
            format: format,
            last_bill_time: fanout_job[:last_bill_time], # Extracted back out of the context in the DerivedAssetProcessor
            first_bill_time: fanout_job[:first_bill_time],
            partition_lower_bound: fanout_job[:partition_lower_bound],
            partition_upper_bound: fanout_job[:partition_upper_bound],
            collection_job_finish_time: collection_job_finish_time,
            data: fanout_job[:resources],
        }

        changed
        notify_observers payload_account_id, payload # See Gdas::Collection::DerivedAssetCollectionTask
      end

      def format_for_service_and_asset(cloud, service_asset_type)
        # Build this lookup once.
        @asset_type_to_format ||=
          ::Gdas::Assets::GenericDerivedAssetScanner::Registry.instance.each_key.with_object({}) do |vals, h|
            hash_key = [vals[:cloud], vals[:service], vals[:asset_type]]
            if h.key?(hash_key)
              # G-DAS assets can register multiple product names, creating multiple entries in this lookup.
              # Because they will have the exact same configurations, just skip the extra registrations.
              logger.debug do
                "Multiple G-DAS registrations for asset type #{hash_key}. "\
                'The "format" value will be used from the first registration.'
              end
              next
            end

            h[hash_key] = vals[:format]
          end

        key = [cloud, service_asset_type[0], service_asset_type[1]]
        @asset_type_to_format[key]
      end
    end

    class Task < ::Gdas::Collection::DerivedAssetCollectionTask
      collector Collector

      def support_caching?
        false
      end
    end
  end
end

# :nocov: Development only
if ImportRegistry.resolve('runtime').sandbox?
  def gdas_context_with_paths(customer_id, cloud, bill_account_id, gdas_xform_s3_path, force, **args)
    context = Hash[
      customer_id: customer_id, cloud: cloud, bill_account_id: bill_account_id,
      gdas_xform_s3_path: gdas_xform_s3_path,
      sla: 'gdas', force: force
    ]
    context = context.merge(args) if args
    context
  end

  def collect_generic_derived_asset_scanner(customer_id, cloud, bill_account_id, gdas_xform_s3_path, force: true)
    collect_from(
      Gdas::Assets::GenericDerivedAssetScanner::Scanner::Collector.new,
      gdas_context_with_paths(customer_id, cloud, bill_account_id, gdas_xform_s3_path, force)
    )
  end

  def task_generic_derived_asset_scanner(customer_id, cloud, bill_account_id, gdas_xform_s3_path, force: true)
    collect_from(
      Gdas::Assets::GenericDerivedAssetScanner::Scanner::Task.new,
      gdas_context_with_paths(customer_id, cloud, bill_account_id, gdas_xform_s3_path, force)
    )
  end
end
# :nocov:
