# frozen_string_literal: true
module Gdas
  module Collection
    # DerivedCollectors allow a new asset to hook into the generic derived asset scanner.
    # The asset should define a collector that extends this class and calls +register_derived_collector+.
    # That's all the setup needed.
    #
    # The class methods below such as gdas_artifact_name, job_partition_column, etc. can be overridden by the
    # new collector if required. For standard G-DAS assets the defaults are okay, but if adding custom fields,
    # some may need different values.
    #
    # Examples:
    #   register_derived_collector(
    #     service: service,
    #     asset: asset,
    #     product_name: 'Amazon Simple Queue Service',
    #     resource_id_pattern: /(.*?):sqs:([^:]+):([[:digit:]]+):(.*)/,
    #     format: 1
    #   )
    class DerivedCollector < ::CloudHealth::Collection::Collector
      SHA1_MIN = "0000000000000000000000000000000000000000"
      SHA1_MAX = "ffffffffffffffffffffffffffffffffffffffff"

      class << self
        # Register this collector with the generic derived asset scanner based on the provided details.
        #
        # Params:
        # - +service+:: String. The service of the asset. Ex: 'sqs' (Simple Queue Service)
        # - +asset+:: String. The asset name. Ex: 'queue'
        # - +product_name+:: String or Array. The value in the "ProductName" column of the DBR for this asset.
        #                      This is used to filter the DBR, and must be exact. Strings will be converted
        #                      to one-element lists.
        # - +resource_id_pattern+:: (Optional) RegEx. If there are multiple assets with the same product
        #                             name, this regular expression will be matched against the
        #                             "ResourceId" column of the bill to match the asset.
        # - +format+:: (Optional) FixNum. The format number of the collection payload. This is usually 1,
        #                and is defaulted to 1. This is helpful when there are multiple collectors for the
        #                same asset.
        def register_derived_collector(
          service: nil,
          asset: nil,
          product_name: nil,
          product_match_name: nil,
          resource_id_pattern: nil,
          format: 1
        )
          raise "service can't be blank for #{name}" if service.blank?
          raise "asset can't be blank for #{name}" if asset.blank?
          raise "product_name can't be blank for #{name}" if product_name.blank? && product_match_name.blank?

          key = {
            service: service,
            asset_type: asset,
            product_name: Array(product_name), # Force to an array. Strings will be wrapped in [].
            product_match_name: product_match_name,
            resource_id_pattern: resource_id_pattern,
            format: format,
            cloud: cloud_name
          }

          # Add key to the registry. If the key is already defined by a different class, raise an error.
          # Due to the way assets and collectors are initialized, it is possible that the same class
          # attempts to register multiple times.
          if self.registry.instance.key?(key)
            previous_registration = self.registry.instance[key]

            return if previous_registration == self
            raise "#{key} already defined by #{previous_registration.name}"
          end
          self.registry.instance[key] = self
        end

        # The name of the G-DAS xform artifact that contains the data to derive this asset type
        def gdas_artifact_name
          raise NotImplementedError
        end

        # The name of the column found in the G-DAS artifact that will be used for processing job partition and
        # cache loading
        def job_partition_column
          'ResourceIdSha1'
        end

        # The minimum value of the job partition field
        def partition_val_min
          SHA1_MIN
        end

        # The maximum value of the job partition field
        def partition_val_max
          SHA1_MAX
        end

        # A function for incrementing the partition value to the next value in the series
        def increment_partition_val(val)
          val == SHA1_MAX ? SHA1_MAX : "%040x" % (val.to_i(16) + 1)
        end

        # The registry that registered asset types will be entered in
        def registry
          ::Gdas::Assets::GenericDerivedAssetScanner::Registry
        end

        def cloud_name
          raise NotImplementedError
        end
      end
    end
  end
end
