# frozen_string_literal: true

module Gdas
  module Consumers
    class GenericDerivedAssetsConsumer < ::CloudHealth::Collection::Consumers::BaseConsumer
      include Import['logger', 'alert']

      class_attribute :account_attr

      def consume
        # Example of params:
        # {
        #   customer_id: 2017,
        #   cloud: 'aws'
        #   bill_account_id: 3367254377911,
        #   gdas_xform_s3_path: 's3://cht-faster-cubes/xform/2017/3367254377911/2018-05/gdas-aggregate-report/gdas-artifact.csv.zip'
        # }

        begin
          # TODO: Remove after classic deploy to change xform kafka payload has gone out
          # Maintain backwards compatibility until then
          params['cloud'] ||= 'aws'
          if params['aws_account_id'].present? && params['bill_account_id'].blank?
            params['bill_account_id'] = params['aws_account_id']
          end

          validate_params!

          # See Gdas::Workers::ProcessingAssetsWorker
          params['sla'] = 'gdas'

          scanner_task = Gdas::Assets::GenericDerivedAssetScanner::Scanner::Task.new
          # The Kafka dispatcher in cht-lib-fabric requires that this be a hash, not a Kafka Params object.
          scanner_task.perform params.to_h.with_indifferent_access
        rescue
          set_logging_context
          raise
        end
      rescue ::Platform::CustomerNotFoundError
        logger.error "Customer not found: #{params['customer_id']}"
      rescue ::Platform::AccountNotFoundError
        logger.error(
          "Account not found: #{params['bill_account_id']} for customer #{params['customer_id']}"
        )
      rescue ::Platform::NonRetryableError => ex
        # A NonRetryableError without an `original_exception` is a way to short circuit out of code
        # and in those cases, don't log or rollbar since that code already would have
        return unless ex.original_exception
        logger.error "#{ex.class}: #{ex.original_exception.message}"
        alert.exception(
          ex,
          customer_id: params['customer_id'],
          bill_account_id: params['bill_account_id'],
          message: ex.original_exception.message
        )
      rescue => ex
        logger.error "#{ex.class}: #{ex.message}"
        alert.exception(
          ex,
          customer_id: params['customer_id'],
          bill_account_id: params['bill_account_id']
        )
        raise # reraise to let sidekiq/karafka handle re-queuing
      end

      def set_logging_context
        ::Logging.mdc.context['customer_id'] = params['customer_id']
        ::Logging.mdc.context['bill_account_id'] = params['bill_account_id']
      end

      def validate_params!
        raise ::Platform::CustomerNotFoundError, params if params['customer_id'].to_i.zero?
        raise ::Platform::AccountNotFoundError, params if params['bill_account_id'].to_i.zero?
        if params['gdas_xform_s3_path'].blank?
          message = 'Required S3 path missing from message for generic derived asset scanner.'
          logger.error "#{message} gdas_xform_s3_path: '#{params['gdas_xform_s3_path']}', Customer: #{params['customer_id']}"
          alert.error(
            message,
            customer_id: params['customer_id'],
            cloud: params['cloud'],
            bill_account_id: params['bill_account_id'],
            gdas_xform_s3_path: params['gdas_xform_s3_path']
          )
          raise ::Platform::NonRetryableError
        end
      end
    end
  end
end
