module Gdas::Processing::Aws
  class AwsDerivedAssetProcessor < ::Gdas::Processing::DerivedAssetProcessor
    include ::Gdas::Helpers::RegionHelper

    # Data source values for processor registrations.
    AWS_GDAS_BILL = 'aws-gdas-bill'.freeze

    # Asset identifiers are a standard set for generic derived assets
    asset_identifiers(
      :aws_region_id,
      :resource_id,
      { db: :resource_id_sha1, sha_source: :resource_id } # The sha1 is used in a SQL index to help lookups.
    )

    self.account_attr = :aws_account_id # database column on asset tables that references the account

    class << self
      def account_model
        ::AwsAccount
      end

      def cloud_name
        ::Gdas::Clouds::AwsCloudConfig::CLOUD_NAME
      end

      def job_partition_field
        'resource_id_sha1'
      end

      def resource_id_sha1_source_field
        :resource_id
      end

      def region_id_field
        :aws_region_id
      end
    end

    def generate_partition_field(asset)
      ::Digest::SHA1.hexdigest asset[self.class.resource_id_sha1_source_field]
    end

    def tag_cache_query(table)
      ::AwsTag
        .select(%w(aws_tags.id `key` `value` `customer_tag`))
        .joins(table.to_sym)
        .where(customer_id: @customer_id, customer_tag: nil)
        .where(table => {
          customer_id: @customer_id,
          is_active: true,
          aws_account_id: @ch_account_id,
          is_derived: true
        })
        .where(partition_clauses)
    end

    # If no asset was found in +lookup_asset+, try looking up the same asset but in the +us-east-1+ region.
    # There was a long-standing bug in G-DAS that caused assets to be incorrectly derived into that region
    # when the bill had minimal information. Now the G-DAS uses the CUR's region field, but without this extra
    # lookup the G-DAS won't find the old asset and instead tries to create a new one, resulting in a duplicate
    # asset error. There is a uniqueness constraint on account + resource_id on these asset tables.
    # See also: asset_identifiers above.
    def additional_asset_lookup(identifier, attrs)
      region_identifier_index = self.class.model_identifiers.find_index(self.class.region_id_field)
      if region_identifier_index && identifier.is_a?(Array) && identifier[region_identifier_index] != 3 # 3 = us-east-1
        identifier_dup = identifier.dup
        identifier_dup[region_identifier_index] = 3 # 3 = us-east-1
        asset = @assets[identifier_dup]
        unless asset
          # Asset wasn't in the cache, so try a DB call to make sure. This covers the case of an asset that
          # has both the wrong region _and_ the wrong inactivated_at, causing it to miss the cache.
          attrs_dup = attrs.dup
          attrs_dup[self.class.region_id_field] = 3 # 3 = us-east-1
          asset = self.class.klass_model
                      .unscoped
                      .where(@lookup_attributes)
                      .where(attrs_dup)
                      .first
        end

        if asset
          logger.warn "[G-DAS lookup_asset] Reverted to us-east-1 region lookup for: #{identifier}"
          asset.send :"#{self.class.region_id_field}=", identifier[region_identifier_index] # Set the region to the new value.
          return asset
        end
      end
      return nil
    end

    def region_from_region_name(region_name)
      aws_region_from_region_name(region_name)
    end
    memoize :region_from_region_name # Memoized per region name

    def log_resurrection(asset, most_recently_found_time)
      logger.warn "Resurrection will cause reingest for customer #{asset.customer_id} and billing account #{asset.aws_account.payer_account_id}: #{asset.inspect}, #{most_recently_found_time}"
    end
  end
end
