# frozen_string_literal: true

module Gdas::Processing::Aws
  class AwsLifetimeProcessor < ::Gdas::Processing::LifetimeProcessor
    include ::Gdas::Helpers::RegionHelper

    set_callback :process_assets, :before, :set_identifiers_in_payload

    AWS_GDAS_BILL = 'aws-gdas-bill'.freeze
    # No real consistency among AWS lifetime assets, each will define this in their registration
    # If the uniqueness of the asset involves a creation timestamp, include it in the `timestamp_identifier`
    # method, but not here.
    # asset_identifiers()

    self.account_attr = :aws_account_id # database column on asset tables that references the account

    class << self
      # The db model representing accounts that own assets
      def account_model
        ::AwsAccount
      end

      def cloud_name
        ::Gdas::Clouds::AwsCloudConfig::CLOUD_NAME
      end

      # resource_id, table_name, instance_id, etc.
      def primary_identifier
        raise NotImplementedError
      end
    end

    # Query to load all the assets that appear in this job. Batch loads assets, using the index on
    # customer_id + primary_identifier.
    def base_asset_queries
      identifiers = data.map { |resource_data| get_identifier(resource_data['resource_id']) }
      queries = []
      identifiers.each_slice(1000) do |chunk|
        query = self.class
                    .klass_model
                    .unscoped
                    .where(@lookup_attributes)
                    .where(self.class.primary_identifier => chunk)

        queries << query
      end
      queries
    end

    # Overriding from `AssetProcessor` in `cht-framework-cp` to support multiple assets that have same `resource_id` when building cache
    def build_cache(query, identifiers, value_prop = nil)
      # Use read shard for these intensive queries
      ::Multidb.use(@customer.read_shard) do
        query.each_with_object(Hash.new{ |h, k| h[k] = [] }) do |row, h|
          id = identifiers.map do |prop|
            val = prop.is_a?(::Hash) ? row.send(prop[:db]) : row.send(prop)
            val.is_a?(::Time) ? self.class.format_time(val) : val
          end
          row = value_prop.call(row) unless value_prop.blank?
          raise 'build_cache generated an empty identifier' if id.blank?
          id = id.first if id.length == 1
          h[id] << row
        end
      end
    end

    # Overriding from `common.rb` to support how we track IDs for multiple assets that have same `resource_id`
    def load_assets_cache
      timer = ::Hitimes::Interval.now
      built_cache = {}
      @ids = ::Set.new
      base_asset_queries.each do |query|
        cache = build_cache(query, self.class.model_identifiers)

        built_cache.merge!(cache)
      end

      # Track the IDs of each record in the cache.
      built_cache.each do |_, assets|
        assets.each do |asset|
          @ids.add(asset.id)
        end
      end

      duration = timer.stop * 1000 # Convert the metric to milliseconds
      tags = %W(
        service:#{@options.dig(:payload, :service)}
        asset:#{@options.dig(:payload, :asset)}
      )
      statsd.timing "gdas.#{processor_type}.asset_cache_load_time", duration, tags: tags

      logger.debug("Asset cache size: #{built_cache.size}")
      built_cache
    end

    # Overriding to support multiple assets that have same `resource_id` and looking up the correct one based on timestamps
    def base_asset_lookup(dso)
      identifier, bad_prop = dso_to_identifier dso, self.class.model_identifiers

      assets = @assets[identifier]
      asset = find_asset_from_overlaps(dso, assets)
      return asset if asset

      primary_attrs = dso_to_attrs dso, self.class.model_identifiers
      asset = attrs_lookup(dso, primary_attrs)
      return asset if asset

      if self.class.model_fallback_ids
        fallback_attrs = dso_to_attrs dso, self.class.model_fallback_ids
        asset = attrs_lookup(dso, fallback_attrs)
        return asset if asset
      end

      asset = additional_asset_lookup(identifier, primary_attrs)
      return asset if asset

      # Throw skip if no asset was found from the lookup
      throw :skip
    end

    # Find asset based on lookups with given attrs (can be primary or fallbacks)
    def attrs_lookup(dso, attrs)
      query = self.class.klass_model
                  .unscoped
                  .where(@lookup_attributes)
                  .where(attrs)

      find_asset_from_overlaps(dso, query)
    end

    # From overlaps between `start_time` and `most_recently_found_time` either choose:
    # (1) If there is a single overlap, choose the asset that overlaps
    # (3) If there are multiple overlaps, choose the asset that is more recent
    # (4) If there is no overlap, choose the asset that is more recent from `start_time`
    def find_asset_from_overlaps(dso, assets)
      return nil if assets.blank?

      # Order by newest to oldest based on `updated_at`
      assets = assets.sort_by(&:updated_at).reverse

      start_time, most_recently_found_time = parse_asset_times(dso)
      prev = assets.first
      assets.each do |asset|
        # Case with overlaps
        return asset if (most_recently_found_time >= asset.created_at && start_time <= asset.updated_at)
        # Case without overlaps
        return prev if start_time > asset.updated_at
        prev = asset
      end
      prev
    end

    # Extracting region as required as well as extracting the primary identifier (e.g. extract name from the resource ID)
    def set_identifiers_in_payload
      data.reject! do |dso|
        if self.class.model_identifiers.include?(:aws_region_id)
          region = aws_region_from_region_name(dso[:region])
          return true unless region
          dso[:aws_region_id] = region.id
        end

        identifier = get_identifier(dso['resource_id'])
        return true unless identifier
        dso[self.class.primary_identifier] = identifier
        false
      end
    end

    # Given the resource id of an asset, calculate its primary identifier (e.g. name)
    def get_identifier(resource_id)
      raise NotImplementedError
    end
  end
end
