# frozen_string_literal: true

module Gdas::Processing::Azure

  # Subclasses of this processor for each Azure asset need to implement the following functions:
  # - +static_resource_id_strs+  used in calculating the hidden_unique_id
  class AzureLifetimeProcessor < ::Gdas::Processing::LifetimeProcessor

    # Data source for processor registrations.
    AZURE_GDAS_BILL = 'azure-gdas-bill'.freeze

    SUBSCRIPTIONS_STR = '/subscriptions/'
    RESOURCE_GROUPS_STR = '/resourcegroups/'

    asset_identifiers(
      :resource_id
    )
    self.account_attr = :azure_subscription_id # database column on asset tables that references the account

    class << self
      # The db model representing accounts that own assets
      def account_model
        ::AzureSubscription
      end

      def cloud_name
        ::Gdas::Clouds::AzureCloudConfig::CLOUD_NAME
      end
    end

    # Query to load all the assets that appear in this job. Batch loads assets, using the index on
    # customer_id + hidden_unique_id.
    def base_asset_queries
      hidden_unique_ids = data.map { |resource_data| get_unique_id(resource_data['resource_id']) }
      queries = []
      hidden_unique_ids.each_slice(1000) do |chunk|
        queries << self.class
                       .klass_model
                       .unscoped
                       .where(@lookup_attributes)
                       .where(hidden_unique_id: chunk)
      end
      queries
    end

    def base_asset_lookup(dso)
      identifier, bad_prop = dso_to_identifier(dso, self.class.model_identifiers)
      asset = @assets[identifier]

      return asset, nil if asset

      attrs = { hidden_unique_id: get_unique_id(dso['resource_id']) }
      asset = self.class.klass_model
                  .unscoped
                  .where(@lookup_attributes)
                  .where(attrs)
                  .first
      return asset, attrs
    end

    # Overridden from framework to downcase keys
    def build_cache(query, identifiers, value_prop = nil)
      # Use read shard for these intensive queries
      ::Multidb.use(@customer.read_shard) do
        query.each_with_object({}) do |row, h|
          id = identifiers.map do |prop|
            val = prop.is_a?(::Hash) ? row.send(prop[:db]) : row.send(prop)
            val.is_a?(::Time) ? self.class.format_time(val) : val
          end
          row = value_prop.call(row) unless value_prop.blank?
          raise 'build_cache generated an empty identifier' if id.blank?
          id = id.first if id.length == 1

          # This line is the only difference for azure since azure bills have case inconsistencies
          id = id.downcase if id.is_a?(String)

          h[id] = row
        end
      end
    end

    # Given the resource id of an asset, calculate its hidden_unique_id
    def get_unique_id(resource_id)
      unique_id = resource_id
      unique_id = unique_id.gsub(/#{SUBSCRIPTIONS_STR}/, '').gsub(/#{RESOURCE_GROUPS_STR}/, '')

      unless static_resource_id_strs.nil?
        static_resource_id_strs.each do |value|
          if value.downcase.include? "classic"
            unique_id = unique_id.sub(/#{value}/, 'classic')
          else
            unique_id = unique_id.sub(/#{value}/, '')
          end
        end
      end
      if unique_id.size > 255
        Digest::SHA2.new(512).hexdigest(unique_id)
      else
        unique_id
      end
    end

    # A list of static substrings that appear in the resource_id for an asset type. This is different for each
    # asset type, see: https://github.com/CloudHealth/cht_azure/blob/35f90bcc50301e3d342cd4a6cb5e878578426834/lib/cht_azure/utils/unique_id_helper.rb#L93
    def static_resource_id_strs
      raise NotImplementedError
    end
  end
end
