module Gdas::Processing
  module Common
    extend ::ActiveSupport::Concern

    included do
      include Import['cache.redis', 'cache.common_assets', 'runtime', 'statsd']
      include ::Gdas::Context::Account

      self.account_func = :ch_account

      def start_timer
        @processing_timer = ::Hitimes::Interval.now
      end

      def stop_timer
        duration = @processing_timer.stop * 1000 # Convert the metric to milliseconds

        tags = %W(
        service:#{@options.dig(:payload, :service)}
        asset:#{@options.dig(:payload, :asset)}
        )

        statsd.timing "gdas.#{processor_type}.time", duration, tags: tags
      end

      def log_job_details(action)
        lower_bound, upper_bound = get_partition_bounds

        logger.warn do
          "#{action} G-DAS #{processor_type} job for customer: #{@options.dig(:customer_id)}, "\
        "account: #{@options.dig(:ch_account_id)}, and "\
        "asset type: #{@options.dig(:payload, :service)} #{@options.dig(:payload, :asset)}, "\
        "from xform path: #{@options.dig(:gdas_xform_s3_path)}, "\
        "containing #{@options.dig(:payload, :data)&.size} data elements, "\
        "with job partition range: #{lower_bound} - #{upper_bound}"
        end
      end

      def get_partition_bounds
        lower_bound = @options.dig(:payload, :partition_lower_bound)
        upper_bound = @options.dig(:payload, :partition_upper_bound)

        [lower_bound, upper_bound]
      end

      def validate_collection_results!(results)
        nil # GDAS does not support account validation
      end

      # Create a cache of the current derived asset. See +base_asset_queries+ for more information.
      # This is distinct from the +AssetProcessor+ version because it handles multiple queries used to build a
      # cache, merging the results together.
      def load_assets_cache
        timer = ::Hitimes::Interval.now
        built_cache = {}
        @ids = ::Set.new
        base_asset_queries.each do |query|
          cache = build_cache(query, self.class.model_identifiers)

          built_cache.merge!(cache)
        end

        # Track the IDs of each record in the cache.
        built_cache.values.each { |v| @ids.add(v.id) }

        duration = timer.stop * 1000 # Convert the metric to milliseconds
        tags = %W(
        service:#{@options.dig(:payload, :service)}
        asset:#{@options.dig(:payload, :asset)}
        )
        statsd.timing "gdas.#{processor_type}.asset_cache_load_time", duration, tags: tags

        logger.debug("Asset cache size: #{built_cache.size}")
        built_cache
      end

      # Get the start and most recent times for this asset, if an error occurs parsing, skip this asset
      def parse_asset_times(dso)
        begin
          most_recently_found_time = Time.parse(dso[:most_recently_found_time])
          start_time = Time.parse(dso[:start_time])
          return start_time, most_recently_found_time
        rescue ArgumentError
          message = 'Failed to convert asset time string to a time.'
          logger.error(
            "#{message} Service: #{@options.dig(:payload, :service)}, Asset: #{@options.dig(:payload, :asset)}"
          )
          alert.error(
            message,
            cloud: @options.dig(:payload, :cloud),
            asset_data: dso,
            customer_id: @customer_id
          )
          throw :skip # Just skip this asset so the rest can still be processed.
        end
      end

      # Attempt to lookup a preexisting asset in the database. Return the asset (if found) and any attributes extracted
      # from the dso.
      def base_asset_lookup(dso)
        identifier, bad_prop = dso_to_identifier dso, self.class.model_identifiers
        asset = @assets[identifier]
        return asset, nil if asset

        attrs = dso_to_attrs dso, self.class.model_identifiers
        asset = self.class.klass_model
                    .unscoped
                    .where(@lookup_attributes)
                    .where(attrs)
                    .first
        return asset, attrs if asset

        asset = additional_asset_lookup(identifier, attrs)
        return asset, attrs if asset

        # This is standard from the +AssetProcessor+.

        # if we have at least one property with a bad value, call skip.
        # we already issued a rollbar in dso_to_identifier
        throw :skip unless bad_prop.nil?
        return nil, attrs
      end

      # Override this to add any additional strategies for looking up an asset
      def additional_asset_lookup(identifier, attrs)
        return nil
      end

      # Update the asset timestamps (`created_at`, `inactivated_at`, `updated_at`) based on the bill times. `inactivated_at`
      # is only set if the asset has already been inactivated through collection, and `updated_at` is automatically set
      # through ActiveRecord.
      def update_asset_timestamps(asset, dso)
        start_time, most_recently_found_time = parse_asset_times(dso)

        if start_time < asset.created_at
          @cross_month_adjustments += 1 if start_time.month != asset.created_at.month
          asset.created_at = start_time
        end

        unless asset.is_active
          asset.inactivated_at = most_recently_found_time if asset.inactivated_at.nil? || most_recently_found_time > asset.inactivated_at
        end
      end

    end
  end
end
