# frozen_string_literal: true

require 'time'
require 'set'
require 'hitimes'

module Gdas::Processing

  # Cloud specific subclasses of this processor need to:
  # - call +asset_identifiers+ at the top level
  # - set +self.account_attr+ to the appropriate database column on asset tables
  # and implement the following functions:
  # - +account_model+
  # - +cloud_name+
  # - +job_partition_field+  db column that processing jobs are partitioned by
  # - +region_id_field+  db column for region id
  # - +generate_partition_field+  create the +job_partition_field+ for an asset
  # - +tag_cache_query+  generate the tag cache corresponding to an asset table
  # - +region_from_region_name+  get a region model based on its name in the dso
  #
  # They can also optionally override these methods to introduce additional logic:
  # - +additional_asset_lookup+  any additional strategies to find an asset from +lookup_asset+
  # - +custom_fields+  can also be used to handle any additional standard fields on the payload

  class DerivedAssetProcessor < ::CloudHealth::Processing::AssetProcessor
    include ::Gdas::Processing::Common # from ::Gdas::Processing::AssetProcessor
    include ::Gdas::Processing::Tags

    set_callback :prepare, :before, -> { start_timer }
    # The +first_bill_time+ is used in the asset cache lookup, so this has to happen before +prepare+.
    set_callback :prepare, :before, :extract_payload_times

    set_callback :process_assets, :before, :set_region_in_payload
    set_callback :process_assets, :before, :load_gdas_tag_cache

    set_callback :execute, :after, -> { stop_timer }

    # +invoke+ is the highest level available. It is defined in +::Tasks::Executor+.
    set_callback :invoke, :before, -> { log_job_details('Starting') }
    set_callback :invoke, :after, -> { log_job_details('Stopping') }

    class << self
      # The db model representing accounts that own assets
      def account_model
        raise NotImplementedError
      end

      # The db model field that should be used for scoping cache loading queries to the given job partition values.
      # Override in derived processor class definition to change partition field.
      def job_partition_field
        raise NotImplementedError
      end

      def updated_at_date_field
        'updated_at_date'
      end

      # Number of hours difference between the last bill time found and the last active timestamp of an asset to
      # declare it inactive
      def gap_for_declaring_inactive
        24
      end

      # Name of the column for region id on asset tables
      def region_id_field
        raise NotImplementedError
      end
    end

    def processor_type
      "derived_processor_#{@options.dig(:payload, :cloud)}"
    end

    # Override in derived processor class definition to support custom fields for an asset type.
    # Should return a hash with the following structure:
    # {
    #   db_field_name1: -> (asset, dso, job_options) { 'do something' },
    #   db_field_name2: -> (asset, dso, job_options) { 'do something else' },
    #   ...
    # }
    def custom_fields
      {}
    end

    # Same format as standard custom fields, but these will be updated regardless of how recent the current row is.
    # This is non-standard behavior, use caution when overriding this function.
    def forced_custom_fields
      {}
    end

    def initialize_metric_counts
      # Created, updated, and deleted are tracked by the audit log.
      @resurrected_assets = 0
      @cross_month_adjustments = 0
      @resurrection_causes_reingest = 0
    end

    def post_metric_counts
      tags = %W(
        cloud:#{self.class.cloud_name}
        service:#{@options.dig(:payload, :service)}
        asset:#{@options.dig(:payload, :asset)}
      )
      post_count(@resurrected_assets, 'resurrected', tags)
      post_count(@cross_month_adjustments, 'cross_month_adjustments', tags)
      post_count(@resurrection_causes_reingest, 'resurrection_causing_reingest', tags)
    end

    def post_count(count, metric_name, tags)
      if count.positive?
        statsd.increment "gdas.assets.#{metric_name}", tags: tags, by: count
      end
    end

    # For each asset in the payload, extract the region from the +usage_type+ and set +aws_region_id+
    # accordingly.
    def set_region_in_payload
      # Track failures to avoid a Rollbar storm. This hash is keyed by the method of extraction.
      # Example: @region_extraction_failures[:region_name] = ['us-west-2-lax-1']
      @region_extraction_failures = {}

      logger.warn "Extracting regions from #{data.length} data elements."

      # Extract the region from the record where possible. If the extraction fails, remove the asset before
      # processing and record an error.
      data.reject! do |dso|
        raise if dso[self.class.region_id_field] # Region shouldn't already exist.

        region = extract_region_from_record(dso)

        if region
          dso[self.class.region_id_field] = region
          false # Good value. Keep in +data+ for processing.
        else
          # Region couldn't be processed. If there was an error, it should have been recorded already.
          # This can happen if a new region is added by Amazon and we haven't added support for it yet.
          # Remove this record from the +data+ sent to processing.
          true
        end
      end

      # If there were errors extracting the region from the record, Rollbar them efficiently.
      @region_extraction_failures.each_pair do |type, errors|
        errors.each do |region_name|
          statsd.increment 'gdas.assets.region_extraction_failure', tags: %W(region_name:#{region_name})
        end

        logger.error(
          'Failed to extract region from generic derived record. '\
          "cloud: #{self.class.cloud_name}, CID: #{@customer_id}, "\
          "extraction method: #{type}, failing examples: #{errors.to_a}"
        )
      end
    end

    # Extract the three times from the context and set instance variables:
    # - Time of the first row seen in the bill
    # - Time of the last row seen in the bill
    # - Time the scan job completed (for setting the +derived_at+ timestamp)
    # This is called before +process_assets+ by the callback above.
    def extract_payload_times
      @first_bill_time = extract_time(:first_bill_time).beginning_of_hour
      @last_bill_time = extract_time(:last_bill_time).beginning_of_hour
      @collection_finish_time = extract_time(:collection_job_finish_time)
    end

    # Extracts the time from the payload designated by the field name. If it does not exist,
    # raise error
    def extract_time(field_name)
      time_int = @options.dig(:payload, field_name)
      unless time_int
        message = "#{field_name} not found in payload."
        logger.error(
            "#{message} Service: #{@options.dig(:payload, :service)}, Asset: #{@options.dig(:payload, :asset)}"
        )
        alert.error(
            message,
            cloud: self.class.cloud_name,
            payload: @options[:payload],
            customer_id: @customer_id
        )
        raise ::Platform::NonRetryableError, orig_exception: message
      end
      Time.at(time_int)
    end

    # Generic derived assets are special. Because the bill may contain assets that have already been
    # inactivated, the cache should load all, unscoped records assets for the customer. This could be a huge
    # query, so an additional, time-based where clause was added. MySQL couldn't use helpful indexes when the
    # query had an +OR+ in it, so these assets require 2 cache-loading queries. See +base_asset_queries+.
    def base_asset_query
      raise NotImplementedError
    end

    def partition_clauses
      lower_bound, upper_bound = get_partition_bounds

      partition_lower_clause = "`#{self.class.klass_model.table_name}`.#{self.class.job_partition_field} >= '#{lower_bound}'"
      partition_upper_clause = "`#{self.class.klass_model.table_name}`.#{self.class.job_partition_field} <= '#{upper_bound}'"
      partition_not_null_clause = "`#{self.class.klass_model.table_name}`.#{self.class.job_partition_field} IS NOT NULL"

      "#{partition_not_null_clause} AND #{partition_lower_clause} AND #{partition_upper_clause}"
    end

    # Return a list of ActiveRecord queries used to build the asset cache.
    # For generic derived assets, there are multiple queries needed to load the required assets in a performant way.
    def base_asset_queries
      raise unless @first_bill_time # Just in case. This is built and checked in +extract_payload_times+.

      # Explaining the queries:
      # - All active, derived assets: This is standard for most processors, and it is needed to ensure that
      #   any active assets that are not found in the bill are properly inactivated. Scoped by the job partition field
      #   since G-DAS jobs are chunked into a constant size.
      # - Batch load all assets in the payload of this job using an `in` query on the job partition field as the
      #   partition field is unique and indexed in the DB. This prevents any lookups later on for assets not in
      #   the cache
      queries = [
          self.class
              .klass_model
              .where(@lookup_attributes)
              .where(is_derived: true)
              .active
              .where(partition_clauses)
      ]

      partition_ids = data.map { |resource_data| resource_data['job_partition_value'] }
      partition_ids.each_slice(1000) do |chunk|
        queries << self.class
                       .klass_model
                       .unscoped
                       .where(@lookup_attributes)
                       .where(self.class.job_partition_field => chunk)
      end

      queries
    end

    # A modified version of the tag_cache_from_table method defined in service/app/gdas/processing/tags.rb
    # that loads derived asset tags within the job partition range
    def load_gdas_tag_cache
      table = self.class.klass_model.table_name
      # Read shard
      ::Multidb.use(@customer.read_shard) do
        @tag_cache = tag_cache_query(table).each_with_object({}) do |t, h|
          h[[t.key, t.value]] = t
        end
      end
    end

    def tag_cache_query(table)
      raise NotImplementedError
    end

    # Custom version of the asset lookup. This version is significantly simpler than the version in the base
    # +AssetProcessor+. Notably, this version does not make guesses about what other lookups should be
    # attempted (ex. nullable fields, timestamps), and it does not automatically resurrect assets when they
    # have to be manually looked up.
    def lookup_asset(dso)
      asset, attrs = base_asset_lookup(dso)
      return asset if asset

      # TODO: first_or_initialize broken in ruby 2.3.0
      # https://github.com/rails/protected_attributes/issues/53#issuecomment-174062735

      # Falling back to creating a new asset.
      self.class.klass_model.new.tap do |t|
        [@lookup_attributes, attrs].each do |y|
          y.each_pair do |k, v|
            t.send :"#{k}=", v
          end
        end
        t.is_derived = true
      end
    end

    # Extract the asset information from the payload and create/update/inactivate DB records accordingly.
    # At least initially, all generic derived assets can share the same processing logic.
    def process_asset(asset, dso)
      # Perform lifetime adjustments on collected assets using `update_asset_timestamps` from `common.rb`
      unless asset.is_derived
        update_asset_timestamps(asset, dso)
        return
      end

      start_time, most_recently_found_time = parse_asset_times(dso)

      asset_was_active = asset.is_active
      asset_active = asset_is_active?(most_recently_found_time, asset.updated_at, asset_was_active)

      if asset.persisted? && !asset_was_active && asset_active # Resurrection!
        @resurrected_assets += 1

        if will_resurrection_cause_reingest?(asset, most_recently_found_time)
          log_resurrection(asset, most_recently_found_time)
          @resurrection_causes_reingest += 1
        end
      end

      # Reset +inactivated_at+ to +nil+ for active assets; update the timestamp if the asset is
      # inactive and it was found more recently than the persisted +inactivated_at+.
      if asset_active
        asset.inactivated_at = nil
      elsif asset.inactivated_at.nil? || asset.inactivated_at < most_recently_found_time
        asset.inactivated_at = most_recently_found_time
      end

      # Set/correct the asset's timestamps
      if asset.persisted?
        # An earlier start time has been found for this asset, so fix it.
        # This could happen if the bills are being processed out of order.
        asset.created_at = start_time if start_time < asset.created_at

        # This is a temporary change to correct derived assets that were updated by another
        # processor which changed the updated_at without changing the is_derived field.
        # This non_hour_timestamp checks if the updated_at value does not land on an hour
        # boundary and allows the G-DAS to update it if that is the case. This is because timestamps
        # in the bill always fall on an hour boundary. See https://cloudhealthtech.atlassian.net/browse/AWS-11269
        non_hour_timestamp = asset.updated_at.min != 0 || asset.updated_at.sec != 0

        if most_recently_found_time == asset.updated_at
          # Tag update functionality checks if the tags are different before attempting any changes
          # Deals with scenarios where historical CUR is regenerated after customer makes cost allocation
          # tags updates
          update_tags(asset, dso[:tags], update_name: false)
        elsif most_recently_found_time > asset.updated_at || non_hour_timestamp
          if non_hour_timestamp
            tags = %W(
                cloud:#{self.class.cloud_name}
                service:#{@options.dig(:payload, :service)}
                asset:#{@options.dig(:payload, :asset)}
            )
            statsd.increment 'gdas.assets.non_hour_bug_fix', tags: tags
          end

          # Only update the +updated_at+ time and tag values for this asset if it's moving forward.
          asset.updated_at = most_recently_found_time
          update_tags(asset, dso[:tags], update_name: false)
          update_custom_fields(asset, dso)
        end
      else
        # New asset
        asset.created_at = start_time
        asset.updated_at = most_recently_found_time
        update_tags(asset, dso[:tags], update_name: false)
        update_custom_fields(asset, dso)

        # If the asset class expects to have the job partition field stored, then store that, too.
        # This is used when the asset id is too large to fit in the database's index.
        # This should already be set during +lookup_asset+ when the asset wasn't found, so check
        # if the calculation is needed before trying.
        if asset.has_attribute?(self.class.job_partition_field) && asset[self.class.job_partition_field].blank?
          asset.send :"#{self.class.job_partition_field}=", generate_partition_field(asset)
        end
      end

      update_custom_fields(asset, dso, forced_custom_fields)

      # If the asset had any changes (including just bumping the timestamps), update +derived_at+.
      # This timestamp is from the collector so all assets created by one job have the same +derived_at+.
      asset.derived_at = @collection_finish_time if asset.changed?

      # If an asset is resurrected, updated, and re-inactivated, it will look like +is_active+ didn't
      # change and the asset will be automatically reactivated:
      # https://github.com/CloudHealth/cht-framework-cp/blob/master/processing/lib/cloud_health/processing/actions/update_asset.rb#L17
      # This line explicitly indicates that the field has changes to avoid the reset.
      asset.is_active_will_change! if !asset_was_active && asset.changed?

      # If we don't explicitly mark a change, ActiveRecord will bump the +updated_at+ to the current time.
      # This allows us to leave the +updated_at+ alone if the asset is from the future.
      asset.updated_at_will_change! if asset.changed? && !asset.updated_at_changed?

      asset.is_active = asset_active

      # Temporarily updating `aws_region_id` in `process_asset` for assets having duplicate key
      # insert errors due to region mismatch
      if asset.has_attribute?(:aws_region_id)
        asset.aws_region_id = extract_region_from_record(dso)
      end
    end

    def update_custom_fields(asset, dso, fields=custom_fields)
      fields.each do |db_field_name, mapping_function|
        catch(:skip_field_update) do
          value = mapping_function.call(asset, dso, @options)
          asset.send "#{db_field_name}=", value
        end
      end
    end

    def generate_partition_field(asset) end

    # If the asset has been updated AFTER the last time in this bill, use whatever state it was already in.
    # If not, check if the asset appeared in the same hour as the last hour in the bill.
    def asset_is_active?(timestamp, asset_updated_at, asset_was_active)
      raise unless @last_bill_time # Just in case. This is built and checked in +extract_payload_times+.

      return asset_was_active if asset_updated_at.present? && asset_updated_at > @last_bill_time

      timestamp > @last_bill_time - self.class.gap_for_declaring_inactive.hours
    end

    # Extract the region ID from the payload.
    # If the lookup fails, track the error. We probably don't support this region yet.
    def extract_region_from_record(dso)
      region = region_from_region_name(dso[:region])
      unless region
        # Invalid region name, so keep track for a Rollbar later.
        @region_extraction_failures[:region_name] ||= Set.new
        @region_extraction_failures[:region_name] << dso[:region]
        return nil # Assets with this region will be skipped.
      end
      region.id
    end

    # Attempt to extract the region from the given region name.
    #
    # Params:
    # - +region_name+:: String. The region name from the superartifact to match against.
    #
    # Returns:
    # - Region model object if there is a match.
    # - +nil+ if the lookup failed.
    def region_from_region_name(region_name)
      raise NotImplementedError
    end

    # If an active asset exists in the database but isn't found in the bill, it won't be processed by
    # +process_asset+ above. Instead, this helper manually deactivates any remaining assets.
    # If an asset is deactivated in the very last hour of the bill, for example, the asset will appear to
    # still be active in the prior month, but will not appear at all in the next month's bill.
    #
    # The +AssetProcessor+ version of this method sets the +inactivated_at+ timestamp to +Time.now+.
    # This version uses the +updated_at+ timestamp from the asset because that is the last time the asset
    # was seen in a bill.
    def inactivate_terminated_assets
      return if @ids.blank?
      terminate_assets = assets_for_termination
      return if terminate_assets.blank?

      logger.debug "Inactivating #{terminate_assets.length} assets..."
      terminate_assets.each do |asset|
        # This keeps the +updated_at+ timestamp from changing when the asset is saved. For derived assets,
        # +updated_at+ means the last time the asset was seen in the bill, *not* the last time a change for
        # the asset was persisted to the database.
        asset.updated_at_will_change!
        ::CloudHealth::Processing::Actions::TerminateAsset.call asset: asset,
                                                                inactivated_at: asset.updated_at + self.class.gap_for_declaring_inactive.hours,
                                                                audit_log: audit_log,
                                                                account: account
      end
    end

    # Override, to exclude already inactive assets.
    def assets_for_termination
      # Only terminate active assets. This is needed because the base asset query includes inactive assets.
      # Additionally, only terminate assets that appear to be inactive as of the end of this bill. This
      # keeps us from incorrectly inactivating assets when the bills are run out of order. The G-DAS processing
      # cache includes every asset in the payload, some of which may be collected assets. The G-DAS must not
      # modify collected assets, so we exclude them from assets for termination.
      @assets.values.select do |asset|
        asset.is_derived && @ids.include?(asset.id) && asset.is_active && asset.updated_at <= @last_bill_time - self.class.gap_for_declaring_inactive.hours
      end
    end

    # If an asset doesn't appear for even one hour, it will be marked as inactive. The next time this job
    # runs, that asset may be resurrected. This is expected for generic derived assets.
    def resurrection_warning?(_asset)
      false
    end
  end
end
