# frozen_string_literal: true


module Gdas::Processing

  # Cloud specific subclasses of this processor need to:
  # - call +asset_identifiers+ at the top level
  # - set +self.account_attr+ to the appropriate database column on asset tables
  # and implement the functions:
  # - +base_asset_queries+
  # - +account_model+
  # They can also optionally override these methods to introduce additional logic:
  # - +additional_asset_lookup+  any additional strategies to find an asset from +lookup_asset+
  class LifetimeProcessor < ::CloudHealth::Processing::AssetProcessor
    include ::Gdas::Processing::Common # from ::Gdas::Processing::AssetProcessor

    set_callback :prepare, :before, -> { start_timer }
    set_callback :process_assets, :before, :initialize_metric_counts
    set_callback :process_assets, :after, :post_metric_counts
    set_callback :execute, :after, -> { stop_timer }

    # +invoke+ is the highest level available. It is defined in +::Tasks::Executor+.
    set_callback :invoke, :before, -> { log_job_details('Starting') }
    set_callback :invoke, :after, -> { log_job_details('Stopping') }

    class << self
      # The db model representing accounts that own assets
      def account_model
        raise NotImplementedError
      end
    end

    def processor_type
      "lifetime_processor_#{@options.dig(:payload, :cloud)}"
    end

    def initialize_metric_counts
      @skipped_assets = 0
      @updated_assets = 0
      @unchanged_assets = 0
      @cross_month_adjustments = 0
    end

    def post_metric_counts
      tags = %W(
        cloud:#{self.class.cloud_name}
        service:#{@options.dig(:payload, :service)}
        asset:#{@options.dig(:payload, :asset)}
      )
      post_count(@skipped_assets, 'asset_not_found', tags)
      post_count(@updated_assets, 'asset_updated', tags)
      post_count(@unchanged_assets, 'asset_unchanged', tags)
      post_count(@cross_month_adjustments, 'cross_month_adjustments', tags)
    end

    def post_count(count, metric_name, tags)
      if count.positive?
        statsd.increment "gdas.assets.lifetime.#{metric_name}", tags: tags, by: count
      end
    end

    # Query to load all the assets that appear in this job.
    def base_asset_queries
      raise NotImplementedError
    end

    # Custom version of the asset lookup. This version is significantly simpler than the version in the base
    # +AssetProcessor+. Notably, this version does not make guesses about what other lookups should be
    # attempted (ex. nullable fields, timestamps), and it does not automatically resurrect assets or create new assets.
    def lookup_asset(dso)
      asset, _ = base_asset_lookup(dso)
      return asset if asset

      # Can't find preexisting asset, skip it
      @skipped_assets += 1
      throw :skip
    end

    # Override `process_asset` to update the asset timestamps (`created_at`, `inactivated_at`, `updated_at`)
    # using `update_asset_timestamps` from `common.rb`
    def process_asset(asset, dso)
      update_asset_timestamps(asset, dso)

      @updated_assets += 1 if asset.changed?
      @unchanged_assets += 1 unless asset.changed?
    end

    # Overriding from framework as the lifetime processors should never inactivate assets.
    def assets_for_termination
      []
    end
  end
end
