require_relative 'aws_asset_base'

class AwsInstance < AwsAssetBase
  include ::Gdas::Constants::Ec2Os
  include ::Gdas::Helpers::ModelTermination
  include ::Gdas::Helpers::ModelDefaultScopes

  belongs_to :customer
  belongs_to :aws_account
  belongs_to :aws_instance_type
  belongs_to :aws_availability_zone
  belongs_to :aws_image
  has_one :aws_region, through: :aws_availability_zone
  has_one :aws_instance_type_profile, through: :aws_instance_type
  has_one :instance_attribute_hash #only created if they have a config mgmt account, like chef
  has_many :instance_file_systems
  has_many :aws_volumes
  has_and_belongs_to_many :aws_security_groups, join_table: :aws_instances_aws_security_groups
  has_and_belongs_to_many :aws_tags, join_table: :aws_instances_aws_tags

  # Ensures disk capacities (bytes per sec and ops per sec) are populated per
  # instance when saving, if they were not assigned before.
  before_save :populate_disk_capacities

  after_update :update_dependent_objects

  terminated_attributes(
    state: {current: %w(pending
                        running
                        shutting-down
                        stopping
                        stopped),
            final: 'terminated'}
  )

  # We need this because aws_instance_type has a default scope in NG.
  def aws_instance_type
    ::AwsInstanceType.unscoped { super }
  end

  class << self
    def model_type_id
      ModelType::AWS_INSTANCE
    end

    def audit_black_list
      [:has_ansible,
       :has_chef,
       :has_puppet,
       :volume_size_in_gb,
       :ebs_optimized,
       :security_group_id_list,
       :tags,
       :last_perfmon_date,
       :launched_by_principal_id,
       :terminated_by_principal_id,
       :hourly_cost,
       :current_projected_cost,
       :current_reservation_pct,
       :disk_bytes_per_sec_capacity,
       :disk_ops_per_sec_capacity]
    end
  end

  def classic_name
    "Amazon EC2 Instance #{instance_id}"
  end

  def crn
    region = ImportRegistry.resolve('cache.common_assets').zone_id_to_region(aws_availability_zone_id)
    [
      'crn:aws:ec2',
      customer_id,
      aws_account_id,
      id,
      launch_time.to_i,
      region.nil? ? '' : region.name,
      "instance/#{instance_id}"
    ].join(':')
  end

  def is_windows?
    return false if is_linux?
    'windows'.casecmp(self.platform) == 0
  end

  def is_linux?
    self.platform.nil? || ('linux'.casecmp(self.platform) == 0)
  end

  protected
  def update_dependent_objects
    self.saved_changes.each_pair do |field, updates|
      orig = updates.first
      curr = updates.last
      if field == 'is_active' && orig && !curr
        unless instance_attribute_hash.nil?
          instance_attribute_hash.is_active = false
          instance_attribute_hash.save
        end

        instance_file_systems.each {|fs|
          fs.is_active = false
          fs.save
        }
      end
    end
  end

  private

  # Ensures disk capacities (bytes per sec and ops per sec) are populated per
  # instance when saving, if they were not assigned before.
  # Uses values stored in the `cht-amazon-pricing` gem.
  def populate_disk_capacities
    # Skipping this step if attributes are missing (e.g. excluded from the SQL
    # SELECT), or if they are already set, or if there is no instance type.
    # Last check at the end to avoid eagerly loading instance types each time.
    if (self.has_attribute?(:disk_bytes_per_sec_capacity) &&
        self.has_attribute?(:disk_ops_per_sec_capacity) &&
        (!self.disk_bytes_per_sec_capacity || !self.disk_ops_per_sec_capacity) &&
        !self.aws_instance_type.nil?)
      profile = self.aws_instance_type.aws_instance_type_profile
      self.disk_bytes_per_sec_capacity = profile.disk_bytes_per_sec_capacity
      self.disk_ops_per_sec_capacity = profile.disk_ops_per_sec_capacity
    end
  end

end
